from dataclasses import KW_ONLY, dataclass

from typing_extensions import TypeVar


@dataclass
class AttributeIORef:
    """Base for references to define IO for an ``Attribute`` over an API.

    This object acts as a specification of the API that its corresponding
    ``AttributeIO`` should access for a given ``Attribute``. The fields necessary to
    distinguish between different ``Attributes`` is an implementation detail of the IO,
    but some examples are a string to send over a TCP port, or URI within an HTTP
    server.
    """

    # Make fields keyword-only so that child classes can have fields without defaults
    _: KW_ONLY
    update_period: float | None = None


AttributeIORefT = TypeVar(
    "AttributeIORefT", bound=AttributeIORef, default=AttributeIORef, covariant=True
)



import asyncio
from collections.abc import Awaitable, Callable, Coroutine
from typing import Any

import tango
from tango import AttrWriteType, Database, DbDevInfo, DevState, server
from tango.server import Device

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.controller_api import ControllerAPI
from fastcs.cs_methods import CommandCallback

from .options import TangoDSROptions
from .util import (
    cast_from_tango_type,
    cast_to_tango_type,
    get_server_metadata_from_attribute,
    get_server_metadata_from_datatype,
)


def _wrap_updater_fget(
    attr_name: str,
    attribute: AttrR,
    controller_api: ControllerAPI,
) -> Callable[[Any], Any]:
    async def fget(tango_device: Device):
        tango_device.info_stream(f"called fget method: {attr_name}")
        return cast_to_tango_type(attribute.datatype, attribute.get())

    return fget


async def _run_threadsafe_blocking(
    coro: Coroutine[Any, Any, Any], loop: asyncio.AbstractEventLoop
) -> None:
    """
    Wraps a concurrent.futures.Future object as an
    asyncio.Future to make it awaitable and then awaits it
    """
    future = asyncio.run_coroutine_threadsafe(coro, loop)
    await asyncio.wrap_future(future)


def _wrap_updater_fset(
    attr_name: str,
    attribute: AttrW,
    controller_api: ControllerAPI,
    loop: asyncio.AbstractEventLoop,
) -> Callable[[Any, Any], Any]:
    async def fset(tango_device: Device, value):
        tango_device.info_stream(f"called fset method: {attr_name}")
        coro = attribute.put(cast_from_tango_type(attribute.datatype, value))
        await _run_threadsafe_blocking(coro, loop)

    return fset


def _collect_dev_attributes(
    root_controller_api: ControllerAPI, loop: asyncio.AbstractEventLoop
) -> dict[str, Any]:
    collection: dict[str, Any] = {}
    for controller_api in root_controller_api.walk_api():
        path = controller_api.path

        for attr_name, attribute in controller_api.attributes.items():
            attr_name = attr_name.title().replace("_", "")
            d_attr_name = f"{'_'.join(path)}_{attr_name}" if path else attr_name

            match attribute:
                case AttrRW():
                    collection[d_attr_name] = server.attribute(
                        label=d_attr_name,
                        fget=_wrap_updater_fget(attr_name, attribute, controller_api),
                        fset=_wrap_updater_fset(
                            attr_name, attribute, controller_api, loop
                        ),
                        access=AttrWriteType.READ_WRITE,
                        **get_server_metadata_from_attribute(attribute),
                        **get_server_metadata_from_datatype(attribute.datatype),
                    )
                case AttrR():
                    collection[d_attr_name] = server.attribute(
                        label=d_attr_name,
                        access=AttrWriteType.READ,
                        fget=_wrap_updater_fget(attr_name, attribute, controller_api),
                        **get_server_metadata_from_attribute(attribute),
                        **get_server_metadata_from_datatype(attribute.datatype),
                    )
                case AttrW():
                    collection[d_attr_name] = server.attribute(
                        label=d_attr_name,
                        access=AttrWriteType.WRITE,
                        fset=_wrap_updater_fset(
                            attr_name, attribute, controller_api, loop
                        ),
                        **get_server_metadata_from_attribute(attribute),
                        **get_server_metadata_from_datatype(attribute.datatype),
                    )

    return collection


def _wrap_command_f(
    method_name: str,
    method: CommandCallback,
    controller_api: ControllerAPI,
    loop: asyncio.AbstractEventLoop,
) -> Callable[..., Awaitable[None]]:
    async def _dynamic_f(tango_device: Device) -> None:
        tango_device.info_stream(
            f"called {'_'.join(controller_api.path)} f method: {method_name}"
        )

        coro = method()
        await _run_threadsafe_blocking(coro, loop)

    _dynamic_f.__name__ = method_name
    return _dynamic_f


def _collect_dev_commands(
    root_controller_api: ControllerAPI,
    loop: asyncio.AbstractEventLoop,
) -> dict[str, Any]:
    collection: dict[str, Any] = {}
    for controller_api in root_controller_api.walk_api():
        path = controller_api.path

        for name, method in controller_api.command_methods.items():
            cmd_name = name.title().replace("_", "")
            d_cmd_name = f"{'_'.join(path)}_{cmd_name}" if path else cmd_name
            collection[d_cmd_name] = server.command(
                f=_wrap_command_f(d_cmd_name, method.fn, controller_api, loop)
            )

    return collection


def _collect_dev_properties(controller_api: ControllerAPI) -> dict[str, Any]:
    collection: dict[str, Any] = {}
    return collection


def _collect_dev_init(controller_api: ControllerAPI) -> dict[str, Callable]:
    async def init_device(tango_device: Device):
        await server.Device.init_device(tango_device)  # type: ignore
        tango_device.set_state(DevState.ON)

    return {"init_device": init_device}


def _collect_dev_flags(controller_api: ControllerAPI) -> dict[str, Any]:
    collection: dict[str, Any] = {}

    collection["green_mode"] = tango.GreenMode.Asyncio

    return collection


def _collect_dsr_args(options: TangoDSROptions) -> list[str]:
    args = []

    if options.debug:
        args.append("-v4")

    return args


class TangoDSR:
    """For controlling a controller with tango"""

    def __init__(
        self,
        controller_api: ControllerAPI,
        loop: asyncio.AbstractEventLoop,
    ):
        self._controller_api = controller_api
        self._loop = loop
        self.dev_class = self._controller_api.__class__.__name__
        self._device = self._create_device()

    def _create_device(self):
        class_dict: dict = {
            **_collect_dev_attributes(self._controller_api, self._loop),
            **_collect_dev_commands(self._controller_api, self._loop),
            **_collect_dev_properties(self._controller_api),
            **_collect_dev_init(self._controller_api),
            **_collect_dev_flags(self._controller_api),
        }

        class_bases = (server.Device,)
        pytango_class = type(self.dev_class, class_bases, class_dict)
        return pytango_class

    def run(self, options: TangoDSROptions | None = None) -> None:
        if options is None:
            options = TangoDSROptions()

        dsr_args = _collect_dsr_args(options)

        server.run(
            (self._device,),
            [self.dev_class, options.dsr_instance, *dsr_args],
            green_mode=server.GreenMode.Asyncio,
        )


def register_dev(dev_name: str, dev_class: str, dsr_instance: str) -> None:
    """Register a device instance in the tango server."""
    dev_info = DbDevInfo()
    dev_info.name = dev_name
    dev_info._class = dev_class  # noqa: SLF001
    dev_info.server = f"{dev_class}/{dsr_instance}"

    db = Database()
    db.delete_device(dev_name)  # Remove existing device entry
    db.add_device(dev_info)

    # Validate registration by reading
    read_dev_info = db.get_device_info(dev_info.name)

    print("Registered on Tango Database:")
    print(f" - Device: {read_dev_info.name}")
    print(f" - Class: {read_dev_info.class_name}")
    print(f" - Device server: {read_dev_info.ds_full_name}")



import asyncio
from dataclasses import dataclass

import aioserial


class NotOpenedError(Exception):
    """If the serial stream is not opened."""

    pass


@dataclass
class SerialConnectionSettings:
    port: str
    baud: int = 115200


class SerialConnection:
    """A serial connection."""

    def __init__(self):
        self.stream = None
        self._lock = asyncio.Lock()

    async def connect(self, settings: SerialConnectionSettings) -> None:
        self.__stream = aioserial.AioSerial(port=settings.port, baudrate=settings.baud)

    @property
    def _stream(self) -> aioserial.AioSerial:
        if self.__stream is None:
            raise NotOpenedError(
                "Need to call connect() before using SerialConnection."
            )

        return self.__stream

    async def send_command(self, message: bytes) -> None:
        async with self._lock:
            await self._send_message(message)

    async def send_query(self, message: bytes, response_size: int) -> bytes:
        async with self._lock:
            await self._send_message(message)
            return await self._receive_response(response_size)

    async def _send_message(self, message):
        await self._stream.write_async(message)

    async def _receive_response(self, size):
        return await self._stream.read_async(size)

    async def close(self) -> None:
        async with self._lock:
            self._stream.close()
            self.__stream = None



import enum
from typing import Any

import numpy as np
import pytest
from pytest_mock import MockerFixture
from softioc import softioc
from tests.assertable_controller import (
    AssertableControllerAPI,
    MyTestAttributeIORef,
    MyTestController,
)
from tests.util import ColourEnum

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.controller import Controller
from fastcs.controller_api import ControllerAPI
from fastcs.cs_methods import Command
from fastcs.datatypes import Bool, Enum, Float, Int, String, Waveform
from fastcs.exceptions import FastCSError
from fastcs.transport.epics.ca.ioc import (
    EPICS_MAX_NAME_LENGTH,
    EpicsCAIOC,
    _add_attr_pvi_info,
    _add_pvi_info,
    _add_sub_controller_pvi_info,
    _create_and_link_read_pv,
    _create_and_link_write_pv,
    _make_record,
)
from fastcs.transport.epics.ca.transport import EpicsCATransport
from fastcs.transport.epics.ca.util import (
    record_metadata_from_attribute,
    record_metadata_from_datatype,
)

DEVICE = "DEVICE"

SEVENTEEN_VALUES = [str(i) for i in range(1, 18)]


class OnOffStates(enum.IntEnum):
    DISABLED = 0
    ENABLED = 1


@pytest.mark.asyncio
async def test_create_and_link_read_pv(mocker: MockerFixture):
    make_record = mocker.patch("fastcs.transport.epics.ca.ioc._make_record")
    add_attr_pvi_info = mocker.patch("fastcs.transport.epics.ca.ioc._add_attr_pvi_info")
    record = make_record.return_value

    attribute = AttrR(Int())
    attribute.add_on_update_callback = mocker.MagicMock()

    _create_and_link_read_pv("PREFIX", "PV", "attr", attribute)

    make_record.assert_called_once_with("PREFIX:PV", attribute)
    add_attr_pvi_info.assert_called_once_with(record, "PREFIX", "attr", "r")

    # Extract the callback generated and set in the function and call it
    attribute.add_on_update_callback.assert_called_once_with(mocker.ANY)
    record_set_callback = attribute.add_on_update_callback.call_args[0][0]
    await record_set_callback(1)

    record.set.assert_called_once_with(1)


@pytest.mark.parametrize(
    "attribute,record_type,kwargs",
    (
        (AttrR(String()), "longStringIn", {}),
        (
            AttrR(Enum(ColourEnum)),
            "mbbIn",
            {"ZRST": "RED", "ONST": "GREEN", "TWST": "BLUE"},
        ),
        (
            AttrR(Enum(enum.IntEnum("ONOFF_STATES", {"DISABLED": 0, "ENABLED": 1}))),
            "mbbIn",
            {"ZRST": "DISABLED", "ONST": "ENABLED"},
        ),
        (AttrR(Waveform(np.int32, (10,))), "WaveformIn", {}),
    ),
)
def test_make_input_record(
    attribute: AttrR,
    record_type: str,
    kwargs: dict[str, Any],
    mocker: MockerFixture,
):
    builder = mocker.patch("fastcs.transport.epics.ca.util.builder")

    pv = "PV"
    _make_record(pv, attribute)
    kwargs.update(record_metadata_from_datatype(attribute.datatype))
    kwargs.update(record_metadata_from_attribute(attribute))

    getattr(builder, record_type).assert_called_once_with(
        pv,
        **kwargs,
    )


def test_make_record_raises(mocker: MockerFixture):
    # Pass a mock as attribute to provoke the fallback case matching on datatype
    with pytest.raises(FastCSError):
        _make_record("PV", mocker.MagicMock())


@pytest.mark.asyncio
async def test_create_and_link_write_pv(mocker: MockerFixture):
    make_record = mocker.patch("fastcs.transport.epics.ca.ioc._make_record")
    add_attr_pvi_info = mocker.patch("fastcs.transport.epics.ca.ioc._add_attr_pvi_info")
    record = make_record.return_value

    attribute = AttrW(Int())
    attribute.put = mocker.AsyncMock()
    attribute.add_sync_setpoint_callback = mocker.MagicMock()

    _create_and_link_write_pv("PREFIX", "PV", "attr", attribute)

    make_record.assert_called_once_with(
        "PREFIX:PV", attribute, on_update=mocker.ANY, out_record=True
    )
    add_attr_pvi_info.assert_called_once_with(record, "PREFIX", "attr", "w")

    # Extract the write update callback generated and set in the function and call it
    attribute.add_sync_setpoint_callback.assert_called_once_with(mocker.ANY)
    sync_setpoint_callback = attribute.add_sync_setpoint_callback.call_args[0][0]
    await sync_setpoint_callback(1)

    record.set.assert_called_once_with(1, process=False)

    # Extract the on update callback generated and set in the function and call it
    on_update_callback = make_record.call_args[1]["on_update"]
    await on_update_callback(1)

    attribute.put.assert_called_once_with(1)


class LongEnum(enum.Enum):
    THIS = 0
    IS = 1
    AN = 2
    ENUM = 3
    WITH = 4
    ALTOGETHER = 5
    TOO = 6
    MANY = 7
    VALUES = 8
    TO = 9
    BE = 10
    DESCRIBED = 11
    BY = 12
    MBB = 14
    TYPE = 15
    EPICS = 16
    RECORDS = 17


@pytest.mark.parametrize(
    "attribute,record_type,kwargs",
    (
        (
            AttrW(Enum(enum.IntEnum("ONOFF_STATES", {"DISABLED": 0, "ENABLED": 1}))),
            "mbbOut",
            {"ZRST": "DISABLED", "ONST": "ENABLED"},
        ),
    ),
)
def test_make_output_record(
    attribute: AttrW,
    record_type: str,
    kwargs: dict[str, Any],
    mocker: MockerFixture,
):
    builder = mocker.patch("fastcs.transport.epics.ca.util.builder")
    update = mocker.MagicMock()

    pv = "PV"
    _make_record(pv, attribute, on_update=update, out_record=True)

    kwargs.update(record_metadata_from_datatype(attribute.datatype, out_record=True))
    kwargs.update(record_metadata_from_attribute(attribute))
    kwargs.update({"always_update": True, "on_update": update})

    getattr(builder, record_type).assert_called_once_with(
        pv,
        **kwargs,
    )


def test_long_enum_validator(mocker: MockerFixture):
    builder = mocker.patch("fastcs.transport.epics.ca.util.builder")
    update = mocker.MagicMock()
    attribute = AttrRW(Enum(LongEnum))
    pv = "PV"
    record = _make_record(pv, attribute, on_update=update, out_record=True)
    validator = builder.longStringOut.call_args.kwargs["validate"]
    assert validator(record, "THIS")  # value is one of the Enum names
    assert not validator(record, "an invalid string value")


def test_get_output_record_raises(mocker: MockerFixture):
    # Pass a mock as attribute to provoke the fallback case matching on datatype
    with pytest.raises(FastCSError):
        _make_record("PV", mocker.MagicMock(), on_update=mocker.MagicMock())


class EpicsController(MyTestController):
    read_int = AttrR(Int(), io_ref=MyTestAttributeIORef())
    read_write_int = AttrRW(Int(), io_ref=MyTestAttributeIORef())
    read_write_float = AttrRW(Float())
    read_bool = AttrR(Bool())
    write_bool = AttrW(Bool(), io_ref=MyTestAttributeIORef())
    read_string = AttrRW(String())
    enum = AttrRW(Enum(enum.IntEnum("Enum", {"RED": 0, "GREEN": 1, "BLUE": 2})))
    one_d_waveform = AttrRW(Waveform(np.int32, (10,)))


@pytest.fixture()
def epics_controller_api(class_mocker: MockerFixture):
    return AssertableControllerAPI(EpicsController(), class_mocker)


def test_ioc(mocker: MockerFixture, epics_controller_api: ControllerAPI):
    ioc_builder = mocker.patch("fastcs.transport.epics.ca.ioc.builder")
    builder = mocker.patch("fastcs.transport.epics.ca.util.builder")
    add_pvi_info = mocker.patch("fastcs.transport.epics.ca.ioc._add_pvi_info")
    add_sub_controller_pvi_info = mocker.patch(
        "fastcs.transport.epics.ca.ioc._add_sub_controller_pvi_info"
    )

    EpicsCAIOC(DEVICE, epics_controller_api)

    # Check records are created
    builder.boolIn.assert_called_once_with(
        f"{DEVICE}:ReadBool",
        **record_metadata_from_attribute(epics_controller_api.attributes["read_bool"]),
        **record_metadata_from_datatype(
            epics_controller_api.attributes["read_bool"].datatype
        ),
    )
    builder.longIn.assert_any_call(
        f"{DEVICE}:ReadInt",
        **record_metadata_from_attribute(epics_controller_api.attributes["read_int"]),
        **record_metadata_from_datatype(
            epics_controller_api.attributes["read_int"].datatype
        ),
    )
    builder.aIn.assert_called_once_with(
        f"{DEVICE}:ReadWriteFloat_RBV",
        **record_metadata_from_attribute(
            epics_controller_api.attributes["read_write_float"]
        ),
        **record_metadata_from_datatype(
            epics_controller_api.attributes["read_write_float"].datatype
        ),
    )
    builder.aOut.assert_any_call(
        f"{DEVICE}:ReadWriteFloat",
        always_update=True,
        on_update=mocker.ANY,
        **record_metadata_from_attribute(
            epics_controller_api.attributes["read_write_float"]
        ),
        **record_metadata_from_datatype(
            epics_controller_api.attributes["read_write_float"].datatype,
            out_record=True,
        ),
    )
    builder.longIn.assert_any_call(
        f"{DEVICE}:ReadWriteInt_RBV",
        **record_metadata_from_attribute(
            epics_controller_api.attributes["read_write_int"]
        ),
        **record_metadata_from_datatype(
            epics_controller_api.attributes["read_write_int"].datatype
        ),
    )
    builder.longOut.assert_called_with(
        f"{DEVICE}:ReadWriteInt",
        always_update=True,
        on_update=mocker.ANY,
        **record_metadata_from_attribute(
            epics_controller_api.attributes["read_write_int"]
        ),
        **record_metadata_from_datatype(
            epics_controller_api.attributes["read_write_int"].datatype, out_record=True
        ),
    )
    builder.mbbIn.assert_called_once_with(
        f"{DEVICE}:Enum_RBV",
        **record_metadata_from_attribute(epics_controller_api.attributes["enum"]),
        **record_metadata_from_datatype(
            epics_controller_api.attributes["enum"].datatype
        ),
    )
    builder.mbbOut.assert_called_once_with(
        f"{DEVICE}:Enum",
        always_update=True,
        on_update=mocker.ANY,
        **record_metadata_from_attribute(epics_controller_api.attributes["enum"]),
        **record_metadata_from_datatype(
            epics_controller_api.attributes["enum"].datatype, out_record=True
        ),
    )
    builder.boolOut.assert_called_once_with(
        f"{DEVICE}:WriteBool",
        always_update=True,
        on_update=mocker.ANY,
        **record_metadata_from_attribute(epics_controller_api.attributes["write_bool"]),
        **record_metadata_from_datatype(
            epics_controller_api.attributes["write_bool"].datatype, out_record=True
        ),
    )
    ioc_builder.Action.assert_any_call(
        f"{DEVICE}:Go", on_update=mocker.ANY, blocking=True
    )

    # Check info tags are added
    add_pvi_info.assert_called_once_with(f"{DEVICE}:PVI")
    add_sub_controller_pvi_info.assert_called_once_with(DEVICE, epics_controller_api)


def test_add_pvi_info(mocker: MockerFixture):
    builder = mocker.patch("fastcs.transport.epics.ca.ioc.builder")
    controller = mocker.MagicMock()
    controller.path = []
    child = mocker.MagicMock()
    child.path = ["Child"]
    controller.get_sub_controllers.return_value = {"d": child}

    _add_pvi_info(f"{DEVICE}:PVI")

    builder.longStringIn.assert_called_once_with(
        f"{DEVICE}:PVI_PV",
        initial_value=f"{DEVICE}:PVI",
        DESC="The records in this controller",
    )
    record = builder.longStringIn.return_value
    record.add_info.assert_called_once_with(
        "Q:group",
        {
            f"{DEVICE}:PVI": {
                "+id": "epics:nt/NTPVI:1.0",
                "display.description": {"+type": "plain", "+channel": "DESC"},
                "": {"+type": "meta", "+channel": "VAL"},
            }
        },
    )


def test_add_pvi_info_with_parent(mocker: MockerFixture):
    builder = mocker.patch("fastcs.transport.epics.ca.ioc.builder")
    controller = mocker.MagicMock()
    controller.path = []
    child = mocker.MagicMock()
    child.path = ["Child"]
    controller.get_sub_controllers.return_value = {"d": child}

    child = mocker.MagicMock()
    _add_pvi_info(f"{DEVICE}:Child:PVI", f"{DEVICE}:PVI", "child")

    builder.longStringIn.assert_called_once_with(
        f"{DEVICE}:Child:PVI_PV",
        initial_value=f"{DEVICE}:Child:PVI",
        DESC="The records in this controller",
    )
    record = builder.longStringIn.return_value
    record.add_info.assert_called_once_with(
        "Q:group",
        {
            f"{DEVICE}:Child:PVI": {
                "+id": "epics:nt/NTPVI:1.0",
                "display.description": {"+type": "plain", "+channel": "DESC"},
                "": {"+type": "meta", "+channel": "VAL"},
            },
            f"{DEVICE}:PVI": {
                "value.child.d": {
                    "+channel": "VAL",
                    "+type": "plain",
                    "+trigger": "value.child.d",
                }
            },
        },
    )


def test_add_sub_controller_pvi_info(mocker: MockerFixture):
    add_pvi_info = mocker.patch("fastcs.transport.epics.ca.ioc._add_pvi_info")
    parent_api = mocker.MagicMock()
    parent_api.path = []
    child_api = mocker.MagicMock()
    child_api.path = ["Child"]
    parent_api.sub_apis = {"d": child_api}

    _add_sub_controller_pvi_info(DEVICE, parent_api)

    add_pvi_info.assert_called_once_with(
        f"{DEVICE}:Child:PVI", f"{DEVICE}:PVI", "child"
    )


def test_add_attr_pvi_info(mocker: MockerFixture):
    record = mocker.MagicMock()

    _add_attr_pvi_info(record, DEVICE, "attr", "r")

    record.add_info.assert_called_once_with(
        "Q:group",
        {
            f"{DEVICE}:PVI": {
                "value.attr.r": {
                    "+channel": "NAME",
                    "+type": "plain",
                    "+trigger": "value.attr.r",
                }
            }
        },
    )


async def do_nothing(): ...


class ControllerLongNames(Controller):
    attr_r_with_reallyreallyreallyreallyreallyreallyreally_long_name = AttrR(Int())
    attr_rw_with_a_reallyreally_long_name_that_is_too_long_for_rbv = AttrRW(Int())
    attr_rw_short_name = AttrRW(Int())
    command_with_reallyreallyreallyreallyreallyreallyreally_long_name = Command(
        do_nothing
    )
    command_short_name = Command(do_nothing)


def test_long_pv_names_discarded(mocker: MockerFixture):
    ioc_builder = mocker.patch("fastcs.transport.epics.ca.ioc.builder")
    builder = mocker.patch("fastcs.transport.epics.ca.util.builder")
    long_name_controller_api = AssertableControllerAPI(ControllerLongNames(), mocker)
    long_attr_name = "attr_r_with_reallyreallyreallyreallyreallyreallyreally_long_name"
    long_rw_name = "attr_rw_with_a_reallyreally_long_name_that_is_too_long_for_RBV"
    assert long_name_controller_api.attributes["attr_rw_short_name"].enabled
    assert long_name_controller_api.attributes[long_attr_name].enabled
    EpicsCAIOC(DEVICE, long_name_controller_api)
    assert long_name_controller_api.attributes["attr_rw_short_name"].enabled
    assert not long_name_controller_api.attributes[long_attr_name].enabled

    short_pv_name = "attr_rw_short_name".title().replace("_", "")
    builder.longOut.assert_called_once_with(
        f"{DEVICE}:{short_pv_name}",
        always_update=True,
        on_update=mocker.ANY,
        **record_metadata_from_datatype(
            long_name_controller_api.attributes["attr_rw_short_name"].datatype,
            out_record=True,
        ),
        **record_metadata_from_attribute(
            long_name_controller_api.attributes["attr_rw_short_name"]
        ),
    )
    builder.longIn.assert_called_once_with(
        f"{DEVICE}:{short_pv_name}_RBV",
        **record_metadata_from_datatype(
            long_name_controller_api.attributes[
                "attr_rw_with_a_reallyreally_long_name_that_is_too_long_for_rbv"
            ].datatype
        ),
        **record_metadata_from_attribute(
            long_name_controller_api.attributes[
                "attr_rw_with_a_reallyreally_long_name_that_is_too_long_for_rbv"
            ]
        ),
    )

    long_pv_name = long_attr_name.title().replace("_", "")
    with pytest.raises(AssertionError):
        builder.longIn.assert_called_once_with(f"{DEVICE}:{long_pv_name}")

    long_rw_pv_name = long_rw_name.title().replace("_", "")
    # neither the readback nor setpoint PV gets made if the full pv name with _RBV
    # suffix is too long
    assert (
        EPICS_MAX_NAME_LENGTH - 4
        < len(f"{DEVICE}:{long_rw_pv_name}")
        < EPICS_MAX_NAME_LENGTH
    )

    with pytest.raises(AssertionError):
        builder.longOut.assert_called_once_with(
            f"{DEVICE}:{long_rw_pv_name}",
            always_update=True,
            on_update=mocker.ANY,
        )
    with pytest.raises(AssertionError):
        builder.longIn.assert_called_once_with(f"{DEVICE}:{long_rw_pv_name}_RBV")

    assert long_name_controller_api.command_methods["command_short_name"].enabled
    long_command_name = (
        "command_with_reallyreallyreallyreallyreallyreallyreally_long_name"
    )
    assert not long_name_controller_api.command_methods[long_command_name].enabled

    short_command_pv_name = "command_short_name".title().replace("_", "")
    ioc_builder.Action.assert_called_once_with(
        f"{DEVICE}:{short_command_pv_name}", on_update=mocker.ANY, blocking=True
    )
    with pytest.raises(AssertionError):
        long_command_pv_name = long_command_name.title().replace("_", "")
        builder.aOut.assert_called_once_with(
            f"{DEVICE}:{long_command_pv_name}",
            initial_value=0,
            always_update=True,
            on_update=mocker.ANY,
        )


def test_update_datatype(mocker: MockerFixture):
    builder = mocker.patch("fastcs.transport.epics.ca.util.builder")

    pv_name = f"{DEVICE}:Attr"

    attr_r = AttrR(Int())
    record_r = _make_record(pv_name, attr_r)

    builder.longIn.assert_called_once_with(
        pv_name,
        **record_metadata_from_attribute(attr_r),
        **record_metadata_from_datatype(attr_r.datatype),
    )
    record_r.set_field.assert_not_called()
    attr_r.update_datatype(Int(units="m", min_alarm=-3))
    record_r.set_field.assert_any_call("EGU", "m")
    record_r.set_field.assert_any_call("LOPR", -3)

    with pytest.raises(
        ValueError,
        match="Attribute datatype must be of type <class 'fastcs.datatypes.Int'>",
    ):
        attr_r.update_datatype(String())  # type: ignore

    attr_w = AttrW(Int())
    record_w = _make_record(pv_name, attr_w, on_update=mocker.ANY, out_record=True)

    builder.longIn.assert_called_once_with(
        pv_name,
        **record_metadata_from_attribute(attr_w),
        **record_metadata_from_datatype(attr_w.datatype),
    )
    record_w.set_field.assert_not_called()
    attr_w.update_datatype(Int(units="m", min_alarm=-1, min=-3))
    record_w.set_field.assert_any_call("EGU", "m")
    record_w.set_field.assert_any_call("LOPR", -1)
    record_w.set_field.assert_any_call("DRVL", -3)

    with pytest.raises(
        ValueError,
        match="Attribute datatype must be of type <class 'fastcs.datatypes.Int'>",
    ):
        attr_w.update_datatype(String())  # type: ignore


def test_ca_context_contains_softioc_commands(mocker: MockerFixture):
    transport = EpicsCATransport(mocker.MagicMock())

    softioc_commands = {
        command: getattr(softioc, command) for command in softioc.command_names
    }
    # We exclude "exit" from the context
    softioc_commands.pop("exit")

    assert transport.context == softioc_commands



from dataclasses import asdict
from typing import Any

from tango import AttrDataFormat

from fastcs.attributes import Attribute
from fastcs.datatypes import Bool, DataType, Enum, Float, Int, String, T, Waveform

TANGO_ALLOWED_DATATYPES = (Bool, DataType, Enum, Float, Int, String, Waveform)

DATATYPE_FIELD_TO_SERVER_FIELD = {
    "units": "unit",
    "min": "min_value",
    "max": "max_value",
    "min_alarm": "min_alarm",
    "max_alarm": "min_alarm",
}


def get_server_metadata_from_attribute(
    attribute: Attribute[T],
) -> dict[str, Any]:
    """Gets the metadata for a Tango field from an attribute."""
    arguments = {}
    arguments["doc"] = attribute.description if attribute.description else ""
    return arguments


def get_server_metadata_from_datatype(datatype: DataType[T]) -> dict[str, str]:
    """Gets the metadata for a Tango field from a FastCS datatype."""
    arguments = {
        DATATYPE_FIELD_TO_SERVER_FIELD[field]: value
        for field, value in asdict(datatype).items()
        if field in DATATYPE_FIELD_TO_SERVER_FIELD
    }

    dtype = datatype.dtype

    match datatype:
        case Waveform():
            dtype = datatype.array_dtype
            match len(datatype.shape):
                case 1:
                    arguments["max_dim_x"] = datatype.shape[0]
                    arguments["dformat"] = AttrDataFormat.SPECTRUM
                case 2:
                    arguments["max_dim_x"], arguments["max_dim_y"] = datatype.shape
                    arguments["dformat"] = AttrDataFormat.IMAGE
                case _:
                    raise TypeError(
                        f"Unsupported shape {datatype.shape}, Tango supports up "
                        "to 2D arrays"
                    )
        case Float():
            arguments["format"] = f"%.{datatype.prec}"

    arguments["dtype"] = dtype
    for argument, value in arguments.items():
        if value is None:
            arguments[argument] = ""

    return arguments


def cast_to_tango_type(datatype: DataType[T], value: T) -> object:
    """Casts a value from FastCS to tango datatype."""
    match datatype:
        case Enum():
            return datatype.index_of(datatype.validate(value))
        case datatype if issubclass(type(datatype), TANGO_ALLOWED_DATATYPES):
            return datatype.validate(value)
        case _:
            raise ValueError(f"Unsupported datatype {datatype}")


def cast_from_tango_type(datatype: DataType[T], value: object) -> T:
    """Casts a value from tango to FastCS datatype."""
    match datatype:
        case Enum():
            return datatype.validate(datatype.members[value])
        case datatype if issubclass(type(datatype), TANGO_ALLOWED_DATATYPES):
            return datatype.validate(value)  # type: ignore
        case _:
            raise ValueError(f"Unsupported datatype {datatype}")



import asyncio
from dataclasses import dataclass, field

from fastcs.controller_api import ControllerAPI
from fastcs.transport.transport import Transport

from .dsr import TangoDSR, TangoDSROptions


@dataclass
class TangoTransport(Transport):
    """Tango transport."""

    dsr: TangoDSROptions = field(default_factory=TangoDSROptions)

    def connect(
        self,
        controller_api: ControllerAPI,
        loop: asyncio.AbstractEventLoop,
    ):
        self._dsr = TangoDSR(controller_api, loop)

    async def serve(self) -> None:
        coro = asyncio.to_thread(self._dsr.run, self.dsr)
        await coro



import asyncio

from p4p.server import Server, StaticProvider

from fastcs.attributes import Attribute, AttrR, AttrRW, AttrW
from fastcs.controller_api import ControllerAPI
from fastcs.transport.epics.util import controller_pv_prefix
from fastcs.util import snake_to_pascal

from ._pv_handlers import (
    make_command_pv,
    make_shared_read_pv,
    make_shared_write_pv,
)
from .pvi_tree import AccessModeType, PviTree


def _attribute_to_access(attribute: Attribute) -> AccessModeType:
    match attribute:
        case AttrRW():
            return "rw"
        case AttrR():
            return "r"
        case AttrW():
            return "w"
        case _:
            raise ValueError(f"Unknown attribute type {type(attribute)}")


async def parse_attributes(
    root_pv_prefix: str, root_controller_api: ControllerAPI
) -> list[StaticProvider]:
    """Parses `Attribute` s into p4p signals in handlers."""
    pvi_tree = PviTree(root_pv_prefix)
    provider = StaticProvider(root_pv_prefix)

    for controller_api in root_controller_api.walk_api():
        pv_prefix = controller_pv_prefix(root_pv_prefix, controller_api)

        pvi_tree.add_sub_device(pv_prefix, controller_api.description)

        for attr_name, attribute in controller_api.attributes.items():
            full_pv_name = f"{pv_prefix}:{snake_to_pascal(attr_name)}"
            match attribute:
                case AttrRW():
                    attribute_pv = make_shared_write_pv(attribute)
                    attribute_pv_rbv = make_shared_read_pv(attribute)
                    provider.add(f"{full_pv_name}", attribute_pv)
                    provider.add(f"{full_pv_name}_RBV", attribute_pv_rbv)
                    pvi_tree.add_signal(f"{full_pv_name}", "rw")
                case AttrR():
                    attribute_pv = make_shared_read_pv(attribute)
                    provider.add(f"{full_pv_name}", attribute_pv)
                    pvi_tree.add_signal(f"{full_pv_name}", "r")
                case AttrW():
                    attribute_pv = make_shared_write_pv(attribute)
                    provider.add(f"{full_pv_name}", attribute_pv)
                    pvi_tree.add_signal(f"{full_pv_name}", "w")

        for attr_name, method in controller_api.command_methods.items():
            full_pv_name = f"{pv_prefix}:{snake_to_pascal(attr_name)}"
            command_pv = make_command_pv(method.fn)
            provider.add(f"{full_pv_name}", command_pv)
            pvi_tree.add_signal(f"{full_pv_name}", "x")

    return [provider, pvi_tree.make_provider()]


class P4PIOC:
    """A P4P IOC which handles a controller"""

    def __init__(self, pv_prefix: str, controller_api: ControllerAPI):
        self.pv_prefix = pv_prefix
        self.controller_api = controller_api

    async def run(self):
        providers = await parse_attributes(self.pv_prefix, self.controller_api)

        endless_event = asyncio.Event()
        with Server(providers):
            await endless_event.wait()



import asyncio
from dataclasses import dataclass, field

from fastcs.controller_api import ControllerAPI
from fastcs.transport.transport import Transport

from .graphql import GraphQLServer
from .options import GraphQLServerOptions


@dataclass
class GraphQLTransport(Transport):
    """GraphQL transport."""

    graphql: GraphQLServerOptions = field(default_factory=GraphQLServerOptions)

    def connect(
        self,
        controller_api: ControllerAPI,
        loop: asyncio.AbstractEventLoop,
    ):
        self._server = GraphQLServer(controller_api)

    async def serve(self) -> None:
        await self._server.serve(self.graphql)

    def __repr__(self) -> str:
        return f"GraphQLTransport({self.graphql.host}:{self.graphql.port})"



import numpy as np
import pytest
from pvi.device import (
    LED,
    ButtonPanel,
    ComboBox,
    Group,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    SubScreen,
    TextFormat,
    TextRead,
    TextWrite,
    ToggleButton,
)
from tests.util import ColourEnum

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.controller_api import ControllerAPI
from fastcs.datatypes import Bool, Enum, Float, Int, String, Waveform
from fastcs.transport.epics.gui import EpicsGUI


def test_get_pv(controller_api):
    gui = EpicsGUI(controller_api, "DEVICE")

    assert gui._get_pv([], "A") == "DEVICE:A"
    assert gui._get_pv(["B"], "C") == "DEVICE:B:C"
    assert gui._get_pv(["D", "E"], "F") == "DEVICE:D:E:F"


@pytest.mark.parametrize(
    "datatype, widget",
    [
        (Bool(), LED()),
        (Int(), TextRead()),
        (Float(), TextRead()),
        (String(), TextRead(format=TextFormat.string)),
        (Enum(ColourEnum), TextRead(format=TextFormat.string)),
        # (Waveform(array_dtype=np.int32), None),
    ],
)
def test_get_attribute_component_r(datatype, widget, controller_api):
    gui = EpicsGUI(controller_api, "DEVICE")

    assert gui._get_attribute_component([], "Attr", AttrR(datatype)) == SignalR(
        name="Attr", read_pv="Attr", read_widget=widget
    )


@pytest.mark.parametrize(
    "datatype, widget",
    [
        (Bool(), ToggleButton()),
        (Int(), TextWrite()),
        (Float(), TextWrite()),
        (String(), TextWrite(format=TextFormat.string)),
        (Enum(ColourEnum), ComboBox(choices=["RED", "GREEN", "BLUE"])),
    ],
)
def test_get_attribute_component_w(datatype, widget, controller_api):
    gui = EpicsGUI(controller_api, "DEVICE")

    assert gui._get_attribute_component([], "Attr", AttrW(datatype)) == SignalW(
        name="Attr", write_pv="Attr", write_widget=widget
    )


def test_get_attribute_component_none(mocker, controller_api):
    gui = EpicsGUI(controller_api, "DEVICE")

    mocker.patch.object(gui, "_get_read_widget", return_value=None)
    mocker.patch.object(gui, "_get_write_widget", return_value=None)
    assert gui._get_attribute_component([], "Attr", AttrR(Int())) is None
    assert gui._get_attribute_component([], "Attr", AttrW(Int())) is None
    assert gui._get_attribute_component([], "Attr", AttrRW(Int())) is None


def test_get_read_widget_none(controller_api):
    gui = EpicsGUI(controller_api, "DEVICE")
    assert gui._get_read_widget(fastcs_datatype=Waveform(np.int32)) is None


def test_get_write_widget_none(controller_api):
    gui = EpicsGUI(controller_api, "DEVICE")
    assert gui._get_write_widget(fastcs_datatype=Waveform(np.int32)) is None


def test_get_components(controller_api):
    gui = EpicsGUI(controller_api, "DEVICE")

    components = gui.extract_api_components(controller_api)
    assert components == [
        Group(
            name="SubController01",
            layout=SubScreen(labelled=True),
            children=[
                SignalR(
                    name="ReadInt",
                    read_pv="DEVICE:SubController01:ReadInt",
                    read_widget=TextRead(),
                )
            ],
        ),
        Group(
            name="SubController02",
            layout=SubScreen(labelled=True),
            children=[
                SignalR(
                    name="ReadInt",
                    read_pv="DEVICE:SubController02:ReadInt",
                    read_widget=TextRead(),
                )
            ],
        ),
        SignalR(name="ReadBool", read_pv="DEVICE:ReadBool", read_widget=LED()),
        SignalR(
            name="ReadInt",
            read_pv="DEVICE:ReadInt",
            read_widget=TextRead(),
        ),
        SignalRW(
            name="ReadString",
            read_pv="DEVICE:ReadString_RBV",
            write_pv="DEVICE:ReadString",
        ),
        SignalRW(
            name="ReadWriteFloat",
            write_pv="DEVICE:ReadWriteFloat",
            write_widget=TextWrite(),
            read_pv="DEVICE:ReadWriteFloat_RBV",
            read_widget=TextRead(),
        ),
        SignalRW(
            name="ReadWriteInt",
            write_pv="DEVICE:ReadWriteInt",
            write_widget=TextWrite(),
            read_pv="DEVICE:ReadWriteInt_RBV",
            read_widget=TextRead(),
        ),
        SignalW(
            name="WriteBool",
            write_pv="DEVICE:WriteBool",
            write_widget=ToggleButton(),
        ),
        SignalX(
            name="Go",
            write_pv="DEVICE:Go",
            write_widget=ButtonPanel(actions={"Go": "1"}),
            value="1",
        ),
    ]


def test_get_components_none(mocker):
    """Test that if _get_attribute_component returns none it is skipped"""

    controller_api = ControllerAPI()
    gui = EpicsGUI(controller_api, "DEVICE")
    mocker.patch.object(gui, "_get_attribute_component", return_value=None)

    components = gui.extract_api_components(controller_api)

    assert components == []



import logging

from loguru import logger as _logger

from ._graylog import GraylogEndpoint as GraylogEndpoint
from ._graylog import GraylogEnvFields as GraylogEnvFields
from ._graylog import GraylogStaticFields as GraylogStaticFields
from ._graylog import parse_graylog_env_fields as parse_graylog_env_fields
from ._graylog import parse_graylog_static_fields as parse_graylog_static_fields
from ._logging import LogLevel, _configure_logger

logger = _logger.bind(logger_name="fastcs")
"""FastCS logger

This is a singleton logger instance to be used throughout the library and in specific
drivers. This logger uses ``loguru`` as underlying logging library, which enables much
simpler configuration as well as structured logging.

Keyword arguments to log statments will be attached as extra fields on the log record.
These fields are displayed separately in the console output and can used for filtering
and metrics in graylog.

It is best to keep the message short and use extra fields for additional information for
messages to be formatted nicely in the console. To add kwargs to format the message
without them appearing as extra fields, prepend the key with ``_``.

.. code-block:: python

    from fastcs.logging import logger

    logger.info("PV put: {pv} = {value}", pv=pv, value=value)

By default messages will be logged with the name ``fastcs``. Within different modules
and classes it can be useful to override this name. This can be done with the ``bind``
method. To create a module logger with its name

.. code-block:: python

    from fastcs.logging import logger as _logger

    logger = _logger.bind(logger_name=__name__)

or to create a class logger with its name

.. code-block:: python

    self.logger = _logger.bind(logger_name=__class__.__name__)

As standard ``loguru`` supports ``trace`` level monitoring, but it should not be used in
fastcs. Instead there is a ``Tracer`` class for verbose logging with fine-grained
controls that can be enabled by the user at runtime.

Use ``configure_logging`` to re-confi the logger at runtime. For more advanced
controls, configure the ``logger`` singleton directly.

See the ``loguru`` docs for more information: https://loguru.readthedocs.io
"""


def configure_logging(
    level: LogLevel | None = None,
    graylog_endpoint: GraylogEndpoint | None = None,
    graylog_static_fields: GraylogStaticFields | None = None,
    graylog_env_fields: GraylogEnvFields | None = None,
):
    """Configure FastCS logger

    This function can be called at any time to
      - Change the console log level
      - Enable/disable graylog logging

    :param level: Log level to set
    :param graylog_endpoint: Endpoint for graylog logging - '<host>:<port>'
    :param graylog_static_fields: Fields to add to graylog messages with static values
    :param graylog_env_fields: Fields to add to graylog messages from env variables

    """
    global logger

    _configure_logger(
        logger, level, graylog_endpoint, graylog_static_fields, graylog_env_fields
    )


# Configure logger with defaults - INFO level and disabled
configure_logging()


class _StdLoggingInterceptHandler(logging.Handler):
    """A std logging handler to forward messages to loguru with nice formatting."""

    def emit(self, record: logging.LogRecord):
        logger.bind(logger_name=self.name).log(record.levelname, record.getMessage())


def intercept_std_logger(logger_name: str):
    """Intercept std logging messages from the given logger

    To find the correct ``logger_name`` for a message. Add a breakpoint in
    ``logging.Logger.callHandlers``, debug and run until the log message appears as
    ``record.msg``. The logger name will be in ``self.name``.

    :param logger_name: Name of the logger to intercept

    """
    handler = _StdLoggingInterceptHandler()
    handler.set_name(logger_name)

    std_logger = logging.getLogger(logger_name)
    std_logger.handlers = [handler]
    std_logger.propagate = False



from fastcs.controller_api import ControllerAPI
from fastcs.util import snake_to_pascal


def controller_pv_prefix(prefix: str, controller_api: ControllerAPI) -> str:
    return ":".join([prefix] + [snake_to_pascal(node) for node in controller_api.path])



class FastCSError(Exception):
    """Base class for general problems in the running of a FastCS transport."""


class LaunchError(FastCSError):
    """For when there is an error in launching FastCS with the given
    transports and controller.
    """



from dataclasses import asdict

from softioc import builder

from fastcs.attributes import Attribute, AttrR, AttrRW, AttrW
from fastcs.datatypes import Bool, DataType, Enum, Float, Int, String, T, Waveform
from fastcs.exceptions import FastCSError

_MBB_FIELD_PREFIXES = (
    "ZR",
    "ON",
    "TW",
    "TH",
    "FR",
    "FV",
    "SX",
    "SV",
    "EI",
    "NI",
    "TE",
    "EL",
    "TV",
    "TT",
    "FT",
    "FF",
)

MBB_STATE_FIELDS = tuple(f"{p}ST" for p in _MBB_FIELD_PREFIXES)
MBB_VALUE_FIELDS = tuple(f"{p}VL" for p in _MBB_FIELD_PREFIXES)
MBB_MAX_CHOICES = len(_MBB_FIELD_PREFIXES)


EPICS_ALLOWED_DATATYPES = (Bool, DataType, Enum, Float, Int, String, Waveform)

DATATYPE_FIELD_TO_RECORD_FIELD = {
    "prec": "PREC",
    "units": "EGU",
    "min": "DRVL",
    "max": "DRVH",
    "min_alarm": "LOPR",
    "max_alarm": "HOPR",
}


def record_metadata_from_attribute(
    attribute: Attribute[T],
) -> dict[str, str | None]:
    """Converts attributes on the `Attribute` to the
    field name/value in the record metadata."""
    return {"DESC": attribute.description}


def record_metadata_from_datatype(
    datatype: DataType[T], out_record: bool = False
) -> dict[str, str]:
    """Converts attributes on the `DataType` to the
    field name/value in the record metadata."""

    arguments = {
        DATATYPE_FIELD_TO_RECORD_FIELD[field]: value
        for field, value in asdict(datatype).items()
        if field in DATATYPE_FIELD_TO_RECORD_FIELD
    }

    if not out_record:
        # in type records don't have DRVL/DRVH fields
        arguments.pop("DRVL", None)
        arguments.pop("DRVH", None)

    match datatype:
        case Waveform():
            if len(datatype.shape) != 1:
                raise TypeError(
                    f"Unsupported shape {datatype.shape}, the EPICS transport only "
                    "supports to 1D arrays"
                )
            arguments["length"] = datatype.shape[0]
        case Enum():
            if len(datatype.members) <= MBB_MAX_CHOICES:
                state_keys = dict(
                    zip(
                        MBB_STATE_FIELDS,
                        datatype.names,
                        strict=False,
                    )
                )
                arguments.update(state_keys)
            elif out_record:  # no validators for in type records

                def _verify_in_datatype(_, value):
                    return value in datatype.names

                arguments["validate"] = _verify_in_datatype

    return arguments


def cast_from_epics_type(datatype: DataType[T], value: object) -> T:
    """Casts from an EPICS datatype to a FastCS datatype."""
    match datatype:
        case Bool():
            if value == 0:
                return False
            elif value == 1:
                return True
            else:
                raise ValueError(f"Invalid bool value from EPICS record {value}")
        case Enum():
            if len(datatype.members) <= MBB_MAX_CHOICES:
                return datatype.validate(datatype.members[value])
            else:  # enum backed by string record
                return datatype.validate(datatype.enum_cls[value])
        case datatype if issubclass(type(datatype), EPICS_ALLOWED_DATATYPES):
            return datatype.validate(value)  # type: ignore
        case _:
            raise ValueError(f"Unsupported datatype {datatype}")


def cast_to_epics_type(datatype: DataType[T], value: T) -> object:
    """Casts from an attribute's datatype to an EPICS datatype."""
    match datatype:
        case Enum():
            if len(datatype.members) <= MBB_MAX_CHOICES:
                return datatype.index_of(datatype.validate(value))
            else:  # enum backed by string record
                return datatype.validate(value).name
        case datatype if issubclass(type(datatype), EPICS_ALLOWED_DATATYPES):
            return datatype.validate(value)
        case _:
            raise ValueError(f"Unsupported datatype {datatype}")


def builder_callable_from_attribute(
    attribute: AttrR | AttrW | AttrRW, make_in_record: bool
):
    """Returns a callable to make the softioc record from an attribute instance."""
    match attribute.datatype:
        case Bool():
            return builder.boolIn if make_in_record else builder.boolOut
        case Int():
            return builder.longIn if make_in_record else builder.longOut
        case Float():
            return builder.aIn if make_in_record else builder.aOut
        case String():
            return builder.longStringIn if make_in_record else builder.longStringOut
        case Enum():
            if len(attribute.datatype.members) > MBB_MAX_CHOICES:
                return builder.longStringIn if make_in_record else builder.longStringOut
            else:
                return builder.mbbIn if make_in_record else builder.mbbOut
        case Waveform():
            return builder.WaveformIn if make_in_record else builder.WaveformOut
        case _:
            raise FastCSError(
                f"EPICS unsupported datatype on {attribute}: {attribute.datatype}"
            )



import asyncio
from dataclasses import dataclass, field

from fastcs.controller_api import ControllerAPI
from fastcs.transport.transport import Transport

from .options import RestServerOptions
from .rest import RestServer


@dataclass
class RestTransport(Transport):
    """Rest Transport Adapter."""

    rest: RestServerOptions = field(default_factory=RestServerOptions)

    def connect(
        self,
        controller_api: ControllerAPI,
        loop: asyncio.AbstractEventLoop,
    ):
        self._server = RestServer(controller_api)

    async def serve(self) -> None:
        await self._server.serve(self.rest)

    def __repr__(self) -> str:
        return f"RestTransport({self.rest.host}:{self.rest.port})"



import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Literal

from p4p import Type, Value
from p4p.nt.common import alarm, timeStamp
from p4p.server import StaticProvider
from p4p.server.asyncio import SharedPV

from .types import p4p_alarm_states, p4p_timestamp_now

AccessModeType = Literal["r", "w", "rw", "d", "x"]

PviName = str


@dataclass
class _PviSignalInfo:
    """For storing a pv and it's access in pvi parsing."""

    pv: str
    access: AccessModeType


def _pascal_to_snake(name: str) -> str:
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def _pv_to_pvi_name(pv: str) -> tuple[str, int | None]:
    leaf = pv.rsplit(":", maxsplit=1)[-1]
    match = re.search(r"(\d+)$", leaf)
    number = int(match.group(1)) if match else None
    string_without_number = re.sub(r"\d+$", "", leaf)
    return _pascal_to_snake(string_without_number), number


class PviDevice(dict[str, "PviDevice"]):
    """For creating a pvi structure in pva."""

    pv_prefix: str
    description: str | None
    device_signal_info: _PviSignalInfo | None

    def __init__(
        self,
        pv_prefix: str,
        description: str | None = None,
        device_signal_info: _PviSignalInfo | None = None,
    ):
        self.pv_prefix = pv_prefix
        self.description = description
        self.device_signal_info = device_signal_info

    def __missing__(self, key: str) -> "PviDevice":
        new_device = PviDevice(pv_prefix=f"{self.pv_prefix}:{key}")
        self[key] = new_device
        return self[key]

    def get_recursively(self, *args: str) -> "PviDevice":
        d = self
        for arg in args:
            d = d[arg]
        return d

    def _get_signal_infos(self) -> dict[str, _PviSignalInfo]:
        device_signal_infos: dict[str, _PviSignalInfo] = {}

        for sub_device_name, sub_device in self.items():
            if sub_device:
                device_signal_infos[f"{sub_device_name}:PVI"] = _PviSignalInfo(
                    pv=f"{sub_device.pv_prefix}:PVI", access="d"
                )
            if sub_device.device_signal_info:
                device_signal_infos[sub_device_name] = sub_device.device_signal_info

        return device_signal_infos

    def _make_p4p_raw_value(self) -> dict:
        p4p_raw_value = defaultdict(dict)
        for pv_leaf, signal_info in self._get_signal_infos().items():
            stripped_leaf = pv_leaf.rstrip(":PVI")
            is_controller = stripped_leaf != pv_leaf
            pvi_name, number = _pv_to_pvi_name(stripped_leaf or pv_leaf)
            if is_controller and number is not None:
                if signal_info.access not in p4p_raw_value[pvi_name]:
                    p4p_raw_value[pvi_name][signal_info.access] = {}
                p4p_raw_value[pvi_name][signal_info.access][f"v{number}"] = (
                    signal_info.pv
                )
            elif is_controller:
                p4p_raw_value[pvi_name][signal_info.access] = signal_info.pv
            else:
                attr_pvi_name = f"{pvi_name}{'' if number is None else number}"
                p4p_raw_value[attr_pvi_name][signal_info.access] = signal_info.pv

        return p4p_raw_value

    def _make_type_for_raw_value(self, raw_value: dict) -> Type:
        p4p_raw_type = []
        for pvi_group_name, access_to_field in raw_value.items():
            pvi_group_structure = []
            for access, field in access_to_field.items():
                if isinstance(field, str):
                    pvi_group_structure.append((access, "s"))
                elif isinstance(field, dict):
                    pvi_group_structure.append(
                        (
                            access,
                            (
                                "S",
                                None,
                                [(v, "s") for v, _ in field.items()],
                            ),
                        )
                    )

            p4p_raw_type.append(
                (pvi_group_name, ("S", "structure", pvi_group_structure))
            )

        return Type(
            [
                ("alarm", alarm),
                ("timeStamp", timeStamp),
                ("display", ("S", None, [("description", "s")])),
                ("value", ("S", "structure", p4p_raw_type)),
            ]
        )

    def make_p4p_value(self) -> Value:
        display = (
            {"display": {"description": self.description}}
            if self.description is not None
            else {}
        )  # Defined here so the value can be (none)

        raw_value = self._make_p4p_raw_value()
        p4p_type = self._make_type_for_raw_value(raw_value)

        try:
            return Value(
                p4p_type,
                {
                    **p4p_alarm_states(),
                    **p4p_timestamp_now(),
                    **display,
                    "value": raw_value,
                },
            )
        except KeyError as e:
            raise ValueError(f"Failed to create p4p Value from {raw_value}") from e

    def make_provider(
        self,
        provider: StaticProvider | None = None,
    ) -> StaticProvider:
        if provider is None:
            provider = StaticProvider("PVI")

        provider.add(
            f"{self.pv_prefix}:PVI",
            SharedPV(initial=self.make_p4p_value()),
        )

        for sub_device in self.values():
            if sub_device:
                sub_device.make_provider(provider=provider)

        return provider


# TODO: This can be dramatically cleaned up after https://github.com/DiamondLightSource/FastCS/issues/122
class PviTree:
    """For storing pvi structures."""

    def __init__(self, pv_prefix: str):
        self._pvi_tree_root: PviDevice = PviDevice(pv_prefix)

    def add_sub_device(
        self,
        device_pv: str,
        description: str | None,
    ):
        if ":" not in device_pv:
            assert device_pv == self._pvi_tree_root.pv_prefix
            self._pvi_tree_root.description = description
        else:
            self._pvi_tree_root.get_recursively(
                *device_pv.split(":")[1:]  # To remove the prefix
            ).description = description

    def add_signal(
        self,
        attribute_pv: str,
        access: AccessModeType,
    ):
        leaf_device = self._pvi_tree_root.get_recursively(*attribute_pv.split(":")[1:])

        if leaf_device.device_signal_info is not None:
            raise ValueError(f"Tried to add the field '{attribute_pv}' twice.")

        leaf_device.device_signal_info = _PviSignalInfo(pv=attribute_pv, access=access)

    def make_provider(self) -> StaticProvider:
        return self._pvi_tree_root.make_provider()



import asyncio
import enum

import numpy as np
import pytest
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture
from tests.assertable_controller import AssertableControllerAPI, MyTestController

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.controller_api import ControllerAPI
from fastcs.datatypes import Bool, Enum, Float, Int, String, Waveform
from fastcs.transport.rest.transport import RestTransport


class RestController(MyTestController):
    read_int = AttrR(Int())
    read_write_int = AttrRW(Int())
    read_write_float = AttrRW(Float())
    read_bool = AttrR(Bool())
    write_bool = AttrW(Bool())
    read_string = AttrRW(String())
    enum = AttrRW(Enum(enum.IntEnum("Enum", {"RED": 0, "GREEN": 1, "BLUE": 2})))
    one_d_waveform = AttrRW(Waveform(np.int32, (10,)))
    two_d_waveform = AttrRW(Waveform(np.int32, (10, 10)))


@pytest.fixture(scope="class")
def rest_controller_api(class_mocker: MockerFixture):
    return AssertableControllerAPI(RestController(), class_mocker)


def create_test_client(rest_controller_api: ControllerAPI) -> TestClient:
    rest_transport = RestTransport()
    rest_transport.connect(rest_controller_api, asyncio.AbstractEventLoop())
    return TestClient(rest_transport._server._app)


class TestRestServer:
    @pytest.fixture(scope="class")
    def test_client(self, rest_controller_api):
        with create_test_client(rest_controller_api) as test_client:
            yield test_client

    def test_read_write_int(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = 0
        with rest_controller_api.assert_read_here(["read_write_int"]):
            response = test_client.get("/read-write-int")
        assert response.status_code == 200
        assert response.json()["value"] == expect
        new = 9
        with rest_controller_api.assert_write_here(["read_write_int"]):
            response = test_client.put("/read-write-int", json={"value": new})
        assert test_client.get("/read-write-int").json()["value"] == new

    def test_read_int(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = 0
        with rest_controller_api.assert_read_here(["read_int"]):
            response = test_client.get("/read-int")
        assert response.status_code == 200
        assert response.json()["value"] == expect

    def test_read_write_float(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = 0
        with rest_controller_api.assert_read_here(["read_write_float"]):
            response = test_client.get("/read-write-float")
        assert response.status_code == 200
        assert response.json()["value"] == expect
        new = 0.5
        with rest_controller_api.assert_write_here(["read_write_float"]):
            response = test_client.put("/read-write-float", json={"value": new})
        assert test_client.get("/read-write-float").json()["value"] == new

    def test_read_bool(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = False
        with rest_controller_api.assert_read_here(["read_bool"]):
            response = test_client.get("/read-bool")
        assert response.status_code == 200
        assert response.json()["value"] == expect

    def test_write_bool(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        with rest_controller_api.assert_write_here(["write_bool"]):
            test_client.put("/write-bool", json={"value": True})

    def test_enum(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        enum_attr = rest_controller_api.attributes["enum"]
        assert isinstance(enum_attr, AttrRW)
        enum_cls = enum_attr.datatype.dtype
        assert isinstance(enum_attr.get(), enum_cls)
        assert enum_attr.get() == enum_cls(0)
        expect = 0
        with rest_controller_api.assert_read_here(["enum"]):
            response = test_client.get("/enum")
        assert response.status_code == 200
        assert response.json()["value"] == expect
        new = 2
        with rest_controller_api.assert_write_here(["enum"]):
            response = test_client.put("/enum", json={"value": new})
        assert test_client.get("/enum").json()["value"] == new
        assert isinstance(enum_attr.get(), enum_cls)
        assert enum_attr.get() == enum_cls(2)

    def test_1d_waveform(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        attribute = rest_controller_api.attributes["one_d_waveform"]
        expect = np.zeros((10,), dtype=np.int32)
        assert isinstance(attribute, AttrRW)
        assert np.array_equal(attribute.get(), expect)
        assert isinstance(attribute.get(), np.ndarray)

        with rest_controller_api.assert_read_here(["one_d_waveform"]):
            response = test_client.get("one-d-waveform")
        assert np.array_equal(response.json()["value"], expect)
        new = [1, 2, 3]
        with rest_controller_api.assert_write_here(["one_d_waveform"]):
            test_client.put("/one-d-waveform", json={"value": new})
        assert np.array_equal(test_client.get("/one-d-waveform").json()["value"], new)

        result = test_client.get("/one-d-waveform")
        assert np.array_equal(result.json()["value"], new)
        assert np.array_equal(attribute.get(), new)
        assert isinstance(attribute.get(), np.ndarray)

    def test_2d_waveform(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        attribute = rest_controller_api.attributes["two_d_waveform"]
        assert isinstance(attribute, AttrRW)
        expect = np.zeros((10, 10), dtype=np.int32)
        assert np.array_equal(attribute.get(), expect)
        assert isinstance(attribute.get(), np.ndarray)

        with rest_controller_api.assert_read_here(["two_d_waveform"]):
            result = test_client.get("/two-d-waveform")
        assert np.array_equal(result.json()["value"], expect)
        new = [[1, 2, 3], [4, 5, 6]]
        with rest_controller_api.assert_write_here(["two_d_waveform"]):
            test_client.put("/two-d-waveform", json={"value": new})

        result = test_client.get("/two-d-waveform")
        assert np.array_equal(result.json()["value"], new)
        assert np.array_equal(attribute.get(), new)
        assert isinstance(attribute.get(), np.ndarray)

    def test_go(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        with rest_controller_api.assert_execute_here(["go"]):
            response = test_client.put("/go")

        assert response.status_code == 204

    def test_read_child1(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = 0
        with rest_controller_api.assert_read_here(["SubController01", "read_int"]):
            response = test_client.get("/SubController01/read-int")
        assert response.status_code == 200
        assert response.json()["value"] == expect

    def test_read_child2(
        self, rest_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = 0
        with rest_controller_api.assert_read_here(["SubController02", "read_int"]):
            response = test_client.get("/SubController02/read-int")
        assert response.status_code == 200
        assert response.json()["value"] == expect



import io
import multiprocessing
import os
import random
import signal
import string
import subprocess
import sys
import time
from collections.abc import Callable, Generator
from multiprocessing.context import DefaultContext
from pathlib import Path
from typing import Any
from uuid import uuid4

import pytest
from aioca import purge_channel_caches
from softioc import builder

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.control_system import build_controller_api
from fastcs.datatypes import Bool, Float, Int, String
from fastcs.logging import configure_logging, logger
from fastcs.logging._logging import LogLevel
from fastcs.transport.tango.dsr import register_dev
from tests.assertable_controller import MyTestAttributeIORef, MyTestController
from tests.example_p4p_ioc import run as _run_p4p_ioc
from tests.example_softioc import run as _run_softioc


@pytest.fixture(scope="function", autouse=True)
def clear_softioc_records():
    builder.ClearRecords()


class BackendTestController(MyTestController):
    read_int: AttrR = AttrR(Int(), io_ref=MyTestAttributeIORef())
    read_write_int: AttrRW = AttrRW(Int(), io_ref=MyTestAttributeIORef())
    read_write_float: AttrRW = AttrRW(Float())
    read_bool: AttrR = AttrR(Bool())
    write_bool: AttrW = AttrW(Bool(), io_ref=MyTestAttributeIORef())
    read_string: AttrRW = AttrRW(String())


@pytest.fixture
def controller():
    return BackendTestController()


@pytest.fixture
def controller_api(controller):
    return build_controller_api(controller)


DATA_PATH = Path(__file__).parent / "data"


@pytest.fixture
def data() -> Path:
    return DATA_PATH


# Prevent pytest from catching exceptions when debugging in vscode so that break on
# exception works correctly (see: https://github.com/pytest-dev/pytest/issues/7409)
if os.getenv("PYTEST_RAISE", "0") == "1":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call: pytest.CallInfo[Any]):
        if call.excinfo is not None:
            raise call.excinfo.value
        else:
            raise RuntimeError(
                f"{call} has no exception data, an unknown error has occurred"
            )

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo: pytest.ExceptionInfo[Any]):
        raise excinfo.value


PV_PREFIX = "".join(random.choice(string.ascii_lowercase) for _ in range(12))
HERE = Path(os.path.dirname(os.path.abspath(__file__)))


def _run_ioc_as_subprocess(
    pv_prefix: str,
    run_ioc: Callable,
    error_queue: multiprocessing.Queue,
    stdout_queue: multiprocessing.Queue,
):
    configure_logging(LogLevel.TRACE)
    logger.add(print)  # forward log messages to stdout for start up detection in tests

    try:
        from pytest_cov.embed import cleanup_on_sigterm
    except ImportError:
        ...
    else:
        cleanup_on_sigterm()

    class QueueWriter(io.TextIOBase):
        def __init__(self, queue):
            self.queue = queue

        def write(self, s):  # type: ignore
            self.queue.put(s)

    try:
        sys.stdout = QueueWriter(stdout_queue)
        run_ioc(pv_prefix=pv_prefix)

    except Exception as e:
        error_queue.put(e)


def run_ioc_as_subprocess(
    run_ioc: Callable, ctxt: DefaultContext
) -> Generator[tuple[str, multiprocessing.Queue], None, None]:
    ioc_startup_timeout = 10
    ioc_startup_timeout_error = TimeoutError("IOC did not start in time")

    pv_prefix = str(uuid4())
    error_queue = ctxt.Queue()
    stdout_queue = ctxt.Queue()
    process = ctxt.Process(
        target=_run_ioc_as_subprocess,
        args=(pv_prefix, run_ioc, error_queue, stdout_queue),
    )
    process.start()

    try:
        start_time = time.monotonic()
        while True:
            try:
                if "Running IOC" in (
                    stdout_queue.get(timeout=ioc_startup_timeout)  # type: ignore
                ):
                    stdout_queue.get()  # get the newline
                    break
            except Exception as error:
                raise ioc_startup_timeout_error from error
            if time.monotonic() - start_time > ioc_startup_timeout:
                raise ioc_startup_timeout_error

        time.sleep(0.1)
        yield pv_prefix, stdout_queue

    finally:
        # Propogate errors
        if not error_queue.empty():
            raise error_queue.get()

        # close ca caches before the event loop
        purge_channel_caches()

        error_queue.close()
        stdout_queue.close()
        process.terminate()
        process.join(timeout=ioc_startup_timeout)


@pytest.fixture(scope="module")
def p4p_subprocess():
    multiprocessing.set_start_method("forkserver", force=True)
    yield from run_ioc_as_subprocess(_run_p4p_ioc, multiprocessing.get_context())


@pytest.fixture(scope="module")
def softioc_subprocess():
    multiprocessing.set_start_method("spawn", force=True)
    yield from run_ioc_as_subprocess(_run_softioc, multiprocessing.get_context())


@pytest.fixture(scope="session")
def tango_system():
    subprocess.run(
        ["podman", "compose", "-f", HERE / "benchmarking" / "compose.yaml", "up", "-d"],
        check=True,
    )
    yield
    subprocess.run(
        ["podman", "compose", "-f", HERE / "benchmarking" / "compose.yaml", "down"],
        check=True,
    )


@pytest.fixture(scope="session")
def register_device():
    attempts = 10
    sleep = 1

    if not os.getenv("TANGO_HOST"):
        raise RuntimeError("TANGO_HOST not defined")

    for attempt in range(1, attempts + 1):
        try:
            register_dev(
                dev_name="MY/BENCHMARK/DEVICE",
                dev_class="TestController",
                dsr_instance="MY_SERVER_INSTANCE",
            )
            break
        except Exception:
            time.sleep(sleep)
        if attempt == attempts:
            raise TimeoutError("Tango device could not be registered")


@pytest.fixture(scope="session")
def test_controller(tango_system, register_device):
    timeout = 10
    process = subprocess.Popen(
        ["python", HERE / "benchmarking" / "controller.py"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )

    start_time = time.monotonic()
    while "Uvicorn running" not in (
        process.stdout.readline().strip()  # type: ignore
    ):
        if time.monotonic() - start_time > timeout:
            raise TimeoutError("Controller did not start in time")

    # close ca caches before the event loop
    purge_channel_caches()

    # Stop buffer from getting full and blocking the subprocess
    for f in [process.stdin, process.stdout, process.stderr]:
        if f:
            f.close()

    yield process

    process.send_signal(signal.SIGINT)
    process.wait(timeout)



import json
from dataclasses import KW_ONLY, dataclass
from typing import Any, Literal, TypeVar

from pydantic import BaseModel, ConfigDict, ValidationError

from fastcs.attribute_io import AttributeIO
from fastcs.attribute_io_ref import AttributeIORef
from fastcs.attributes import Attribute, AttrR, AttrRW, AttrW
from fastcs.connections import IPConnection, IPConnectionSettings
from fastcs.controller import Controller
from fastcs.datatypes import Bool, DataType, Float, Int, String
from fastcs.launch import FastCS
from fastcs.transport.epics.ca.transport import EpicsCATransport
from fastcs.transport.epics.options import EpicsIOCOptions


class TemperatureControllerParameter(BaseModel):
    model_config = ConfigDict(extra="forbid")

    command: str
    type: Literal["bool", "int", "float", "str"]
    access_mode: Literal["r", "rw"]

    @property
    def fastcs_datatype(self) -> DataType:
        match self.type:
            case "bool":
                return Bool()
            case "int":
                return Int()
            case "float":
                return Float()
            case "str":
                return String()


def create_attributes(parameters: dict[str, Any]) -> dict[str, Attribute]:
    attributes: dict[str, Attribute] = {}
    for name, parameter in parameters.items():
        name = name.replace(" ", "_").lower()

        try:
            parameter = TemperatureControllerParameter.model_validate(parameter)
        except ValidationError as e:
            print(f"Failed to validate parameter '{parameter}'\n{e}")
            continue

        io_ref = TemperatureControllerAttributeIORef(parameter.command)
        match parameter.access_mode:
            case "r":
                attributes[name] = AttrR(parameter.fastcs_datatype, io_ref=io_ref)
            case "rw":
                attributes[name] = AttrRW(parameter.fastcs_datatype, io_ref=io_ref)

    return attributes


NumberT = TypeVar("NumberT", int, float)


@dataclass
class TemperatureControllerAttributeIORef(AttributeIORef):
    name: str
    _: KW_ONLY
    update_period: float | None = 0.2


class TemperatureControllerAttributeIO(
    AttributeIO[NumberT, TemperatureControllerAttributeIORef]
):
    def __init__(self, connection: IPConnection):
        super().__init__()

        self._connection = connection

    async def update(self, attr: AttrR[NumberT, TemperatureControllerAttributeIORef]):
        query = f"{attr.io_ref.name}?"
        response = await self._connection.send_query(f"{query}\r\n")
        value = response.strip("\r\n")

        await attr.update(attr.dtype(value))

    async def send(
        self, attr: AttrW[NumberT, TemperatureControllerAttributeIORef], value: NumberT
    ) -> None:
        command = f"{attr.io_ref.name}={attr.dtype(value)}"
        await self._connection.send_command(f"{command}\r\n")


class TemperatureRampController(Controller):
    def __init__(
        self,
        index: int,
        parameters: dict[str, TemperatureControllerParameter],
        io: TemperatureControllerAttributeIO,
    ):
        self._parameters = parameters
        super().__init__(f"Ramp{index}", ios=[io])

    async def initialise(self):
        self.attributes.update(create_attributes(self._parameters))


class TemperatureController(Controller):
    def __init__(self, settings: IPConnectionSettings):
        self._ip_settings = settings
        self._connection = IPConnection()

        self._io = TemperatureControllerAttributeIO(self._connection)
        super().__init__(ios=[self._io])

    async def connect(self):
        await self._connection.connect(self._ip_settings)

    async def initialise(self):
        await self.connect()

        api = json.loads((await self._connection.send_query("API?\r\n")).strip("\r\n"))

        ramps_api = api.pop("Ramps")
        self.attributes.update(create_attributes(api))

        for idx, ramp_parameters in enumerate(ramps_api):
            ramp_controller = TemperatureRampController(
                idx + 1, ramp_parameters, self._io
            )
            await ramp_controller.initialise()
            self.add_sub_controller(f"Ramp{idx + 1:02d}", ramp_controller)

        await self._connection.close()


epics_ca = EpicsCATransport(ca_ioc=EpicsIOCOptions(pv_prefix="DEMO"))
connection_settings = IPConnectionSettings("localhost", 25565)
fastcs = FastCS(TemperatureController(connection_settings), [epics_ca])


if __name__ == "__main__":
    fastcs.run()



from asyncio import iscoroutinefunction
from collections.abc import Callable, Coroutine
from inspect import Signature, getdoc, signature
from types import MethodType
from typing import Any, Generic, TypeVar

from fastcs.controller import BaseController
from fastcs.tracer import Tracer

from .exceptions import FastCSError

MethodCallback = Callable[..., Coroutine[None, None, None]]
"""Generic base class for all `Controller` methods"""
Controller_T = TypeVar("Controller_T", bound=BaseController)
"""Generic `Controller` class that an unbound method must be called with as `self`"""
UnboundCommandCallback = Callable[[Controller_T], Coroutine[None, None, None]]
"""A Command callback that is unbound and must be called with a `Controller` instance"""
UnboundScanCallback = Callable[[Controller_T], Coroutine[None, None, None]]
"""A Scan callback that is unbound and must be called with a `Controller` instance"""
UnboundPutCallback = Callable[[Controller_T, Any], Coroutine[None, None, None]]
"""A Put callback that is unbound and must be called with a `Controller` instance"""
CommandCallback = Callable[[], Coroutine[None, None, None]]
"""A Command callback that is bound and can be called without `self`"""
ScanCallback = Callable[[], Coroutine[None, None, None]]
"""A Scan callback that is bound and can be called without `self`"""
PutCallback = Callable[[Any], Coroutine[None, None, None]]
"""A Put callback that is bound and can be called without `self`"""


method_not_bound_error = NotImplementedError(
    "Method must be bound to a controller instance to be callable"
)


class Method(Generic[Controller_T], Tracer):
    """Generic base class for all FastCS Controller methods."""

    def __init__(self, fn: MethodCallback, *, group: str | None = None) -> None:
        super().__init__()

        self._docstring = getdoc(fn)

        sig = signature(fn, eval_str=True)
        self._parameters = sig.parameters
        self._return_type = sig.return_annotation
        self._validate(fn)

        self._fn = fn
        self._group = group
        self.enabled = True

    def _validate(self, fn: MethodCallback) -> None:
        if self.return_type not in (None, Signature.empty):
            raise FastCSError("Method return type must be None or empty")

        if not iscoroutinefunction(fn):
            raise FastCSError("Method must be async function")

    @property
    def return_type(self):
        return self._return_type

    @property
    def parameters(self):
        return self._parameters

    @property
    def docstring(self):
        return self._docstring

    @property
    def fn(self):
        return self._fn

    @property
    def group(self):
        return self._group


class Command(Method[BaseController]):
    """A `Controller` `Method` that performs a single action when called.

    This class contains a function that is bound to a specific `Controller` instance and
    is callable outside of the class context, without an explicit `self` parameter.
    Calling an instance of this class will call the bound `Controller` method.
    """

    def __init__(self, fn: CommandCallback, *, group: str | None = None):
        super().__init__(fn, group=group)

    def _validate(self, fn: CommandCallback) -> None:
        super()._validate(fn)

        if not len(self.parameters) == 0:
            raise FastCSError(f"Command method cannot have arguments: {fn}")

    async def __call__(self):
        return await self._fn()


class Scan(Method[BaseController]):
    """A `Controller` `Method` that will be called periodically in the background.

    This class contains a function that is bound to a specific `Controller` instance and
    is callable outside of the class context, without an explicit `self` parameter.
    Calling an instance of this class will call the bound `Controller` method.
    """

    def __init__(self, fn: ScanCallback, period: float):
        super().__init__(fn)

        self._period = period

    @property
    def period(self):
        return self._period

    def _validate(self, fn: ScanCallback) -> None:
        super()._validate(fn)

        if not len(self.parameters) == 0:
            raise FastCSError("Scan method cannot have arguments")

    async def __call__(self):
        return await self._fn()


class UnboundCommand(Method[Controller_T]):
    """A wrapper of an unbound `Controller` method to be bound into a `Command`.

    This generic class stores an unbound `Controller` method - effectively a function
    that takes an instance of a specific `Controller` type (`Controller_T`). Instances
    of this class can be added at `Controller` definition, either manually or with use
    of the `command` wrapper, to register the method to be included in the API of the
    `Controller`. When the `Controller` is instantiated, these instances will be bound
    to the instance, creating a `Command` instance.
    """

    def __init__(
        self, fn: UnboundCommandCallback[Controller_T], *, group: str | None = None
    ) -> None:
        super().__init__(fn, group=group)

    def _validate(self, fn: UnboundCommandCallback[Controller_T]) -> None:
        super()._validate(fn)

        if not len(self.parameters) == 1:
            raise FastCSError("Command method cannot have arguments")

    def bind(self, controller: Controller_T) -> Command:
        return Command(MethodType(self.fn, controller), group=self.group)

    def __call__(self):
        raise method_not_bound_error


class UnboundScan(Method[Controller_T]):
    """A wrapper of an unbound `Controller` method to be bound into a `Scan`.

    This generic class stores an unbound `Controller` method - effectively a function
    that takes an instance of a specific `Controller` type (`Controller_T`). Instances
    of this class can be added at `Controller` definition, either manually or with use
    of the `scan` wrapper, to register the method to be included in the API of the
    `Controller`. When the `Controller` is instantiated, these instances will be bound
    to the instance, creating a `Scan` instance.
    """

    def __init__(self, fn: UnboundScanCallback[Controller_T], period: float) -> None:
        super().__init__(fn)

        self._period = period

    @property
    def period(self):
        return self._period

    def _validate(self, fn: UnboundScanCallback[Controller_T]) -> None:
        super()._validate(fn)

        if not len(self.parameters) == 1:
            raise FastCSError("Scan method cannot have arguments")

    def bind(self, controller: Controller_T) -> Scan:
        return Scan(MethodType(self.fn, controller), self._period)

    def __call__(self):
        raise method_not_bound_error



import asyncio
import inspect
import json
from pathlib import Path
from typing import Annotated, Any, Optional, get_type_hints

import typer
from pydantic import BaseModel, ValidationError, create_model
from ruamel.yaml import YAML

from fastcs import __version__
from fastcs.control_system import FastCS
from fastcs.controller import Controller
from fastcs.exceptions import LaunchError
from fastcs.logging import (
    GraylogEndpoint,
    GraylogEnvFields,
    GraylogStaticFields,
    LogLevel,
    configure_logging,
    parse_graylog_env_fields,
    parse_graylog_static_fields,
)
from fastcs.transport import Transport


def launch(
    controller_class: type[Controller],
    version: str | None = None,
) -> None:
    """
    Serves as an entry point for starting FastCS applications.

    By utilizing type hints in a Controller's __init__ method, this
    function provides a command-line interface to describe and gather the
    required configuration before instantiating the application.

    Args:
        controller_class (type[Controller]): The FastCS Controller to instantiate.
            It must have a type-hinted __init__ method and no more than 2 arguments.
        version (Optional[str]): The version of the FastCS Controller.
            Optional

    Raises:
        LaunchError: If the class's __init__ is not as expected

    Example of the expected Controller implementation:
        class MyController(Controller):
            def __init__(self, my_arg: MyControllerOptions) -> None:
                ...

    Typical usage:
        if __name__ == "__main__":
            launch(MyController)
    """
    _launch(controller_class, version)()


def _launch(
    controller_class: type[Controller],
    version: str | None = None,
) -> typer.Typer:
    fastcs_options = _extract_options_model(controller_class)
    launch_typer = typer.Typer()

    class LaunchContext:
        def __init__(self, controller_class, fastcs_options):
            self.controller_class = controller_class
            self.fastcs_options = fastcs_options

    def version_callback(value: bool):
        if value:
            if version:
                print(f"{controller_class.__name__}: {version}")
            print(f"FastCS: {__version__}")
            raise typer.Exit()

    @launch_typer.callback()
    def main(
        ctx: typer.Context,
        version: Optional[bool] = typer.Option(  # noqa (Optional required for typer)
            None,
            "--version",
            callback=version_callback,
            is_eager=True,
            help=f"Display the {controller_class.__name__} version.",
        ),
    ):
        ctx.obj = LaunchContext(
            controller_class,
            fastcs_options,
        )

    @launch_typer.command(help=f"Produce json schema for a {controller_class.__name__}")
    def schema(ctx: typer.Context):
        system_schema = ctx.obj.fastcs_options.model_json_schema()
        print(json.dumps(system_schema, indent=2))

    @launch_typer.command(help=f"Start up a {controller_class.__name__}")
    def run(
        ctx: typer.Context,
        config: Annotated[
            Path,
            typer.Argument(
                help=f"A yaml file matching the {controller_class.__name__} schema"
            ),
        ],
        log_level: Annotated[
            Optional[LogLevel],  # noqa: UP045
            typer.Option(),
        ] = None,
        graylog_endpoint: Annotated[
            Optional[GraylogEndpoint],  # noqa: UP045
            typer.Option(
                help="Endpoint for graylog logging - '<host>:<port>'",
                parser=GraylogEndpoint.parse_graylog_endpoint,
            ),
        ] = None,
        graylog_static_fields: Annotated[
            Optional[GraylogStaticFields],  # noqa: UP045
            typer.Option(
                help="Fields to add to graylog messages with static values",
                parser=parse_graylog_static_fields,
            ),
        ] = None,
        graylog_env_fields: Annotated[
            Optional[GraylogEnvFields],  # noqa: UP045
            typer.Option(
                help="Fields to add to graylog messages from environment variables",
                parser=parse_graylog_env_fields,
            ),
        ] = None,
    ):
        """
        Start the controller
        """
        configure_logging(
            log_level, graylog_endpoint, graylog_static_fields, graylog_env_fields
        )

        controller_class = ctx.obj.controller_class
        fastcs_options = ctx.obj.fastcs_options

        yaml = YAML(typ="safe")
        options_yaml = yaml.load(config)

        try:
            instance_options = fastcs_options.model_validate(options_yaml)
        except ValidationError as e:
            if any("transport" in error["loc"] for error in json.loads(e.json())):
                raise LaunchError(
                    "Failed to validate transports. "
                    "Are the correct fastcs extras installed? "
                    f"Available transports:\n{Transport.subclasses}",
                ) from e

            raise LaunchError("Failed to validate config") from e

        if hasattr(instance_options, "controller"):
            controller = controller_class(instance_options.controller)
        else:
            controller = controller_class()

        instance = FastCS(
            controller, instance_options.transport, loop=asyncio.get_event_loop()
        )

        instance.run()

    return launch_typer


def _extract_options_model(controller_class: type[Controller]) -> type[BaseModel]:
    sig = inspect.signature(controller_class.__init__)
    args = inspect.getfullargspec(controller_class.__init__)[0]
    if len(args) == 1:
        fastcs_options = create_model(
            f"{controller_class.__name__}",
            transport=(list[Transport.union()], ...),
            __config__={"extra": "forbid"},
        )
    elif len(args) == 2:
        hints = get_type_hints(controller_class.__init__)
        if "return" in hints:
            del hints["return"]
        if hints:
            options_type = list(hints.values())[-1]
        else:
            raise LaunchError(
                f"Expected typehinting in '{controller_class.__name__}"
                f".__init__' but received {sig}. Add a typehint for `{args[-1]}`."
            )
        fastcs_options = create_model(
            f"{controller_class.__name__}",
            controller=(options_type, ...),
            transport=(list[Transport.union()], ...),
            __config__={"extra": "forbid"},
        )
    else:
        raise LaunchError(
            f"Expected no more than 2 arguments for '{controller_class.__name__}"
            f".__init__' but received {len(args)} as `{sig}`"
        )
    return fastcs_options


def get_controller_schema(target: type[Controller]) -> dict[str, Any]:
    """Gets schema for a give controller for serialisation."""
    options_model = _extract_options_model(target)
    target_schema = options_model.model_json_schema()
    return target_schema



import asyncio
import signal
from collections.abc import Coroutine, Sequence
from functools import partial
from typing import Any

from IPython.terminal.embed import InteractiveShellEmbed

from fastcs.controller import BaseController, Controller
from fastcs.controller_api import ControllerAPI
from fastcs.cs_methods import Command, Scan, ScanCallback
from fastcs.exceptions import FastCSError
from fastcs.logging import logger as _fastcs_logger
from fastcs.tracer import Tracer
from fastcs.transport import Transport
from fastcs.util import validate_hinted_attributes

tracer = Tracer(name=__name__)
logger = _fastcs_logger.bind(logger_name=__name__)


class FastCS:
    """Entrypoint for a FastCS application.

    This class takes a ``Controller``, creates asyncio tasks to run its update loops and
    builds its API to serve over the given transports.

    :param: controller: The controller to serve in the control system
    :param: transports: A list of transports to serve the API over
    :param: loop: Optional event loop to run the control system in
    """

    def __init__(
        self,
        controller: Controller,
        transports: Sequence[Transport],
        loop: asyncio.AbstractEventLoop | None = None,
    ):
        self._controller = controller
        self._transports = transports
        self._loop = loop or asyncio.get_event_loop()

        self._scan_coros: list[ScanCallback] = []
        self._initial_coros: list[ScanCallback] = []

        self._scan_tasks: set[asyncio.Task] = set()

    def run(self, interactive: bool = True):
        serve = asyncio.ensure_future(self.serve(interactive=interactive))

        self._loop.add_signal_handler(signal.SIGINT, serve.cancel)
        self._loop.add_signal_handler(signal.SIGTERM, serve.cancel)
        self._loop.run_until_complete(serve)

    async def _run_initial_coros(self):
        for coro in self._initial_coros:
            await coro()

    async def _start_scan_tasks(self):
        self._scan_tasks = {self._loop.create_task(coro()) for coro in self._scan_coros}

        for task in self._scan_tasks:
            task.add_done_callback(self._scan_done)

    def _scan_done(self, task: asyncio.Task):
        try:
            task.result()
        except Exception as e:
            raise FastCSError(
                "Exception raised in scan method of "
                f"{self._controller.__class__.__name__}"
            ) from e

    def _stop_scan_tasks(self):
        for task in self._scan_tasks:
            if not task.done():
                try:
                    task.cancel()
                except (asyncio.CancelledError, RuntimeError):
                    pass
                except Exception as e:
                    raise RuntimeError("Unhandled exception in stop scan tasks") from e

        self._scan_tasks.clear()

    async def serve(self, interactive: bool = True) -> None:
        await self._controller.initialise()
        validate_hinted_attributes(self._controller)
        self._controller.connect_attribute_ios()

        self.controller_api = build_controller_api(self._controller)
        self._scan_coros, self._initial_coros = (
            self.controller_api.get_scan_and_initial_coros()
        )

        context = {
            "controller": self._controller,
            "controller_api": self.controller_api,
            "transports": [
                transport.__class__.__name__ for transport in self._transports
            ],
        }

        coros: list[Coroutine] = []
        for transport in self._transports:
            transport.connect(controller_api=self.controller_api, loop=self._loop)
            coros.append(transport.serve())
            common_context = context.keys() & transport.context.keys()
            if common_context:
                raise RuntimeError(
                    "Duplicate context keys found between "
                    f"current context { ({k: context[k] for k in common_context}) } "
                    f"and {transport.__class__.__name__} context: "
                    f"{ ({k: transport.context[k] for k in common_context}) }"
                )
            context.update(transport.context)

        if interactive:
            coros.append(self._interactive_shell(context))
        else:

            async def block_forever():
                while True:
                    await asyncio.sleep(1)

            coros.append(block_forever())

        logger.info(
            "Starting FastCS",
            controller=self._controller,
            transports=f"[{', '.join(str(t) for t in self._transports)}]",
        )

        await self._controller.connect()
        await self._run_initial_coros()
        await self._start_scan_tasks()

        try:
            await asyncio.gather(*coros)
        except asyncio.CancelledError:
            pass
        except Exception:
            logger.exception("Unhandled exception in serve")
        finally:
            logger.info("Shutting down FastCS")
            self._stop_scan_tasks()
            await self._controller.disconnect()

    async def _interactive_shell(self, context: dict[str, Any]):
        """Spawn interactive shell in another thread and wait for it to complete."""

        def run(coro: Coroutine[None, None, None]):
            """Run coroutine on FastCS event loop from IPython thread."""

            def wrapper():
                asyncio.create_task(coro)

            self._loop.call_soon_threadsafe(wrapper)

        async def interactive_shell(
            context: dict[str, object], stop_event: asyncio.Event
        ):
            """Run interactive shell in a new thread."""
            shell = InteractiveShellEmbed()
            await asyncio.to_thread(partial(shell.mainloop, local_ns=context))

            stop_event.set()

        context["run"] = run

        stop_event = asyncio.Event()
        self._loop.create_task(interactive_shell(context, stop_event))
        await stop_event.wait()

    def __del__(self):
        self._stop_scan_tasks()


def build_controller_api(controller: Controller) -> ControllerAPI:
    return _build_controller_api(controller, [])


def _build_controller_api(controller: BaseController, path: list[str]) -> ControllerAPI:
    scan_methods: dict[str, Scan] = {}
    command_methods: dict[str, Command] = {}
    for attr_name in dir(controller):
        attr = getattr(controller, attr_name)
        match attr:
            case Scan(enabled=True):
                scan_methods[attr_name] = attr
            case Command(enabled=True):
                command_methods[attr_name] = attr
            case _:
                pass

    return ControllerAPI(
        path=path,
        attributes=controller.attributes,
        scan_methods=scan_methods,
        command_methods=command_methods,
        sub_apis={
            name: _build_controller_api(sub_controller, path + [name])
            for name, sub_controller in controller.sub_controllers.items()
        },
        description=controller.description,
    )



import getpass
import os
from enum import StrEnum
from logging import LogRecord
from typing import TYPE_CHECKING

from prompt_toolkit.patch_stdout import StdoutProxy
from pygelf import GelfUdpHandler
from pygelf.gelf import SKIP_LIST

from fastcs.logging._graylog import (
    GraylogEndpoint,
    GraylogEnvFields,
    GraylogStaticFields,
)

if TYPE_CHECKING:
    from loguru import Logger
else:
    from typing import Any

    Logger = Any


def _configure_logger(
    logger: Logger,
    level: "LogLevel | None" = None,
    graylog_endpoint: GraylogEndpoint | None = None,
    graylog_static_fields: GraylogStaticFields | None = None,
    graylog_env_fields: GraylogEnvFields | None = None,
):
    logger.remove()
    logger.add(
        sink=StdoutProxy(raw=True),  # type: ignore
        colorize=True,
        format=format_record,
        level=level or "INFO",
    )

    if graylog_endpoint is not None:
        static_fields = {
            "app_name": "fastcs",
            "username": getpass.getuser(),
            "process_id": os.getpid(),
        }
        if graylog_static_fields is not None:
            static_fields.update(graylog_static_fields)

        logger.add(
            LoguruGelfUdpHandler(
                graylog_endpoint.host,
                graylog_endpoint.port,
                # Include built-in file, line, module, function and logger_name fields
                debug=True,
                # Include these static fields
                static_fields=static_fields,
                # Include configured fields from environment
                additional_env_fields=graylog_env_fields,
                # Include fields added dynamically to log record (log statement kwargs)
                include_extra_fields=True,
            ),
            format="{message}",
            level="INFO",
        )

    if level is None:
        logger.disable("fastcs")
    else:
        logger.enable("fastcs")


def format_record(record) -> str:
    _time = record["time"].strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
    _timezone = record["time"].strftime("%z")
    time = f"{_time}{_timezone}"

    name = record["extra"].pop("logger_name", None) or record["name"]

    sep = "<white>,</white> "
    if "extra" in record:
        extras = [
            format_key_value(k, v)
            for k, v in record["extra"].items()
            if not k.startswith("_")
        ]
        extras = f"{sep.join(extras)}"
    else:
        extras = ""

    return f"""\
<level>[{time} {record["level"].name[0]}]</level> \
{record["message"]:<80} \
<green>[{name}]</green> \
{extras}
{{exception}}\
"""


def format_key_value(k, v):
    return f"<cyan>{k}</cyan><white>=</white><magenta>{v}</magenta>"


class LogLevel(StrEnum):
    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class LoguruGelfUdpHandler(GelfUdpHandler):
    def convert_record_to_gelf(self, record: LogRecord):
        # Use full file path instead of file name
        record.__dict__["filename"] = record.__dict__["pathname"]

        # Pull loguru extra fields into root of record for pygelf to include
        extras = record.__dict__.pop("extra", {})
        for k, v in extras.items():
            if k not in SKIP_LIST and not k.startswith("_"):
                record.__dict__[k] = v

        return super().convert_record_to_gelf(record)



from dataclasses import dataclass
from functools import partial
from typing import Generic, TypeVar

import pytest
from pytest_mock import MockerFixture

from fastcs.attribute_io import AttributeIO
from fastcs.attribute_io_ref import AttributeIORef
from fastcs.attributes import (
    AttrR,
    AttrRW,
    AttrW,
)
from fastcs.controller import Controller
from fastcs.datatypes import Float, Int, String, T

NumberT = TypeVar("NumberT", int, float)


@pytest.mark.asyncio
async def test_attributes():
    device = {"state": "Idle", "number": 1, "count": False}
    ui = {"state": "", "number": 0, "count": False}

    async def update_ui(value, key):
        ui[key] = value

    async def send(_attr, value, key):
        device[key] = value

    async def device_add():
        device["number"] += 1

    attr_r = AttrR(String())
    attr_r.add_on_update_callback(partial(update_ui, key="state"))
    await attr_r.update(device["state"])
    assert ui["state"] == "Idle"

    attr_rw = AttrRW(Int())
    attr_rw._on_put_callback = partial(send, key="number")
    attr_rw.add_sync_setpoint_callback(partial(update_ui, key="number"))
    await attr_rw.put(2, sync_setpoint=True)
    assert device["number"] == 2
    assert ui["number"] == 2


@pytest.mark.asyncio
async def test_attribute_io():
    @dataclass
    class MyAttributeIORef(AttributeIORef):
        cool: int

    class MyAttributeIO(AttributeIO[int, MyAttributeIORef]):
        async def update(self, attr: AttrR[T, MyAttributeIORef]):
            print("I am updating", self.ref_type, attr.io_ref.cool)

    class MyController(Controller):
        my_attr = AttrR(Int(), io_ref=MyAttributeIORef(cool=5))
        your_attr = AttrR(Int(), io_ref=MyAttributeIORef(cool=10))

        def __init__(self):
            super().__init__(ios=[MyAttributeIO()])

    c = MyController()

    class ControllerNoIO(Controller):
        my_attr = AttrR(Int(), io_ref=MyAttributeIORef(cool=5))

    @dataclass
    class OtherAttributeIORef(AttributeIORef):
        not_cool: int

    class MissingIOController(Controller):
        my_attr = AttrR(Int(), io_ref=OtherAttributeIORef(not_cool=5))

    with pytest.raises(ValueError, match="does not have an AttributeIO to handle"):
        controller = MissingIOController()
        controller.connect_attribute_ios()

    await c.initialise()
    c.connect_attribute_ios()
    await c.my_attr.bind_update_callback()()


class DummyConnection:
    def __init__(self):
        self._connected = False
        self._int_value = 5
        self._ro_int_value = 10
        self._float_value = 7.5

    async def connect(self):
        self._connected = True

    async def get(self, uri: str):
        if not self._connected:
            raise TimeoutError("No response from DummyConnection")
        if uri == "config/introspect_api":
            return [
                {
                    "name": "int_parameter",
                    "subsystem": "status",
                    "dtype": "int",
                    "min": 0,
                    "max": 100,
                    "value": self._int_value,
                    "read_only": False,
                },
                {
                    "name": "ro_int_parameter",
                    "subsystem": "status",
                    "dtype": "int",
                    "value": self._ro_int_value,
                    "read_only": True,
                },
                {
                    "name": "float_parameter",
                    "subsystem": "status",
                    "dtype": "float",
                    "max": 1000.0,
                    "value": self._float_value,
                    "read_only": False,
                },
            ]

        # increment after getting
        elif uri == "status/int_parameter":
            value = self._int_value
            self._int_value += 1
        elif uri == "status/ro_int_parameter":
            value = self._ro_int_value
            self._ro_int_value += 1
        elif uri == "status/float_parameter":
            value = self._float_value
            self._float_value += 1
        else:
            raise RuntimeError()
        return value

    async def set(self, uri: str, value: float | int):
        if uri == "status/int_parameter":
            self._int_value = value
        elif uri == "status/ro_int_parameter":
            # don't update read only parameter
            pass
        elif uri == "status/float_parameter":
            self._float_value = value


@pytest.mark.asyncio()
async def test_dynamic_attribute_io_specification():
    @dataclass
    class DemoParameterAttributeIORef(AttributeIORef, Generic[NumberT]):
        name: str
        subsystem: str
        connection: DummyConnection

        @property
        def uri(self):
            return f"{self.subsystem}/{self.name}"

    class DemoParameterAttributeIO(AttributeIO[NumberT, DemoParameterAttributeIORef]):
        async def update(
            self,
            attr: AttrR[NumberT, DemoParameterAttributeIORef],
        ):
            value = await attr.io_ref.connection.get(attr.io_ref.uri)
            await attr.update(value)  # type: ignore

        async def send(
            self,
            attr: AttrW[NumberT, DemoParameterAttributeIORef],
            value: NumberT,
        ) -> None:
            await attr.io_ref.connection.set(attr.io_ref.uri, value)
            if isinstance(attr, AttrRW):
                await self.update(attr)

    class DemoParameterController(Controller):
        ro_int_parameter: AttrR
        int_parameter: AttrRW
        float_parameter: AttrRW  # hint to satisfy pyright

        async def initialise(self):
            self._connection = DummyConnection()
            await self._connection.connect()
            dtype_mapping = {"int": Int, "float": Float}
            example_introspection_response = await self._connection.get(
                "config/introspect_api"
            )
            assert isinstance(example_introspection_response, list)
            for parameter_response in example_introspection_response:
                try:
                    ro = parameter_response["read_only"]
                    ref = DemoParameterAttributeIORef(
                        name=parameter_response["name"],
                        subsystem=parameter_response["subsystem"],
                        connection=self._connection,
                    )
                    attr_class = AttrR if ro else AttrRW
                    attr = attr_class(
                        datatype=dtype_mapping[parameter_response["dtype"]](
                            min=parameter_response.get("min", None),
                            max=parameter_response.get("max", None),
                        ),
                        io_ref=ref,
                        initial_value=parameter_response.get("value", None),
                    )

                    self.attributes[ref.name] = attr
                    setattr(self, ref.name, attr)

                except Exception as e:
                    print(
                        "Exception constructing attribute from parameter response:",
                        parameter_response,
                        e,
                    )

    c = DemoParameterController(ios=[DemoParameterAttributeIO()])
    await c.initialise()
    c.connect_attribute_ios()
    await c.ro_int_parameter.bind_update_callback()()
    assert c.ro_int_parameter.get() == 10
    await c.ro_int_parameter.bind_update_callback()()
    assert c.ro_int_parameter.get() == 11

    await c.int_parameter.put(20)
    assert c.int_parameter.get() == 20


@pytest.mark.asyncio
async def test_attribute_no_io(mocker: MockerFixture):
    class MyController(Controller):
        no_ref = AttrRW(Int())
        base_class_ref = AttrRW(Int(), io_ref=AttributeIORef())

    with pytest.raises(
        ValueError,
        match="MyController does not have an AttributeIO to handle AttributeIORef",
    ):
        c = MyController()
        c.connect_attribute_ios()

    class SimpleAttributeIO(AttributeIO[T, AttributeIORef]):
        async def update(self, attr):
            match attr:
                case AttrR(datatype=Int()):
                    await attr.update(100)

    with pytest.raises(
        RuntimeError, match="More than one AttributeIO class handles AttributeIORef"
    ):
        MyController(ios=[SimpleAttributeIO(), SimpleAttributeIO()])

    # we need to explicitly pass an AttributeIO if we want to handle instances of
    # the AttributeIORef base class
    c = MyController(ios=[SimpleAttributeIO()])
    assert not c.no_ref.has_io_ref()
    assert c.base_class_ref.has_io_ref()

    await c.initialise()
    c.connect_attribute_ios()

    # There is a difference between providing an AttributeIO for the default
    # AttributeIORef class and not specifying the io_ref for an Attribute
    # default callbacks are not provided by AttributeIO subclasses

    sync_setpoint_mock = mocker.AsyncMock()
    c.no_ref.add_sync_setpoint_callback(sync_setpoint_mock)

    await c.no_ref.put(40)
    sync_setpoint_mock.assert_called_once_with(40)  # sync setpoint called on first set
    sync_setpoint_mock.reset_mock()
    await c.no_ref.put(41)  # sync setpoint callback not called without flag
    await c.no_ref.put(42, sync_setpoint=True)
    sync_setpoint_mock.assert_called_once_with(42)

    c2 = MyController(ios=[SimpleAttributeIO()])

    await c2.initialise()
    c2.connect_attribute_ios()

    assert c2.base_class_ref.get() == 0
    await c2.base_class_ref.bind_update_callback()()
    assert c2.base_class_ref.get() == 100



from pvi._format.dls import DLSFormatter  # type: ignore
from pvi.device import (
    LED,
    ButtonPanel,
    CheckBox,
    ComboBox,
    ComponentUnion,
    Device,
    Grid,
    Group,
    ReadWidgetUnion,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    SubScreen,
    TableRead,
    TableWrite,
    TextFormat,
    TextRead,
    TextWrite,
    ToggleButton,
    Tree,
    WriteWidgetUnion,
)
from pydantic import ValidationError

from fastcs.attributes import Attribute, AttrR, AttrRW, AttrW
from fastcs.controller_api import ControllerAPI
from fastcs.cs_methods import Command
from fastcs.datatypes import (
    Bool,
    DataType,
    Enum,
    Float,
    Int,
    String,
    Table,
    Waveform,
)
from fastcs.exceptions import FastCSError
from fastcs.util import numpy_to_fastcs_datatype, snake_to_pascal

from .options import EpicsGUIFormat, EpicsGUIOptions


class EpicsGUI:
    """For creating gui in the EPICS transports."""

    def __init__(self, controller_api: ControllerAPI, pv_prefix: str) -> None:
        self._controller_api = controller_api
        self._pv_prefix = pv_prefix

    def _get_pv(self, attr_path: list[str], name: str):
        attr_prefix = ":".join(
            [self._pv_prefix] + [snake_to_pascal(node) for node in attr_path]
        )
        return f"{attr_prefix}:{snake_to_pascal(name)}"

    def _get_read_widget(self, fastcs_datatype: DataType) -> ReadWidgetUnion | None:
        match fastcs_datatype:
            case Bool():
                return LED()
            case Int() | Float():
                return TextRead()
            case String():
                return TextRead(format=TextFormat.string)
            case Enum():
                return TextRead(format=TextFormat.string)
            case Waveform():
                return None
            case datatype:
                raise FastCSError(f"Unsupported type {type(datatype)}: {datatype}")

    def _get_write_widget(self, fastcs_datatype: DataType) -> WriteWidgetUnion | None:
        match fastcs_datatype:
            case Bool():
                return ToggleButton()
            case Int() | Float():
                return TextWrite()
            case String():
                return TextWrite(format=TextFormat.string)
            case Enum():
                return ComboBox(choices=fastcs_datatype.names)
            case Waveform():
                return None
            case datatype:
                raise FastCSError(f"Unsupported type {type(datatype)}: {datatype}")

    def _get_attribute_component(
        self, attr_path: list[str], name: str, attribute: Attribute
    ) -> SignalR | SignalW | SignalRW | None:
        pv = self._get_pv(attr_path, name)
        name = snake_to_pascal(name)
        match attribute:
            case AttrRW():
                read_widget = self._get_read_widget(attribute.datatype)
                write_widget = self._get_write_widget(attribute.datatype)
                if write_widget is None or read_widget is None:
                    return None
                return SignalRW(
                    name=name,
                    description=attribute.description,
                    write_pv=pv,
                    write_widget=write_widget,
                    read_pv=pv + "_RBV",
                    read_widget=read_widget,
                )
            case AttrR():
                read_widget = self._get_read_widget(attribute.datatype)
                if read_widget is None:
                    return None
                return SignalR(
                    name=name,
                    description=attribute.description,
                    read_pv=pv,
                    read_widget=read_widget,
                )
            case AttrW():
                write_widget = self._get_write_widget(attribute.datatype)
                if write_widget is None:
                    return None
                return SignalW(
                    name=name,
                    description=attribute.description,
                    write_pv=pv,
                    write_widget=write_widget,
                )
            case _:
                raise FastCSError(f"Unsupported attribute type: {type(attribute)}")

    def _get_command_component(self, attr_path: list[str], name: str):
        pv = self._get_pv(attr_path, name)
        name = snake_to_pascal(name)

        return SignalX(
            name=name,
            write_pv=pv,
            value="1",
            write_widget=ButtonPanel(actions={name: "1"}),
        )

    def create_gui(self, options: EpicsGUIOptions | None = None) -> None:
        if options is None:
            options = EpicsGUIOptions()

        if options.file_format is EpicsGUIFormat.edl:
            raise FastCSError("FastCS does not support .edl screens.")

        assert options.output_path.suffix == options.file_format.value
        options.output_path.parent.mkdir(parents=True, exist_ok=True)

        components = self.extract_api_components(self._controller_api)
        device = Device(label=options.title, children=components)

        formatter = DLSFormatter()
        formatter.format(device, options.output_path.resolve())

    def extract_api_components(self, controller_api: ControllerAPI) -> Tree:
        components: Tree = []

        for name, api in controller_api.sub_apis.items():
            components.append(
                Group(
                    name=snake_to_pascal(name),
                    layout=SubScreen(),
                    children=self.extract_api_components(api),
                )
            )

        groups: dict[str, list[ComponentUnion]] = {}
        for attr_name, attribute in controller_api.attributes.items():
            try:
                signal = self._get_attribute_component(
                    controller_api.path,
                    attr_name,
                    attribute,
                )
            except ValidationError as e:
                print(f"Invalid name:\n{e}")
                continue

            if signal is None:
                continue

            match attribute:
                case Attribute(group=group) if group is not None:
                    if group not in groups:
                        groups[group] = []

                    # Remove duplication of group name and signal name
                    signal.name = signal.name.removeprefix(group)

                    groups[group].append(signal)
                case _:
                    components.append(signal)

        for name, command in controller_api.command_methods.items():
            signal = self._get_command_component(controller_api.path, name)

            match command:
                case Command(group=group) if group is not None:
                    if group not in groups:
                        groups[group] = []

                    groups[group].append(signal)
                case _:
                    components.append(signal)

        for name, children in groups.items():
            components.append(Group(name=name, layout=Grid(), children=children))

        return components


class PvaEpicsGUI(EpicsGUI):
    """For creating gui in the PVA EPICS transport."""

    def _get_pv(self, attr_path: list[str], name: str):
        return f"pva://{super()._get_pv(attr_path, name)}"

    def _get_read_widget(self, fastcs_datatype: DataType) -> ReadWidgetUnion | None:
        if isinstance(fastcs_datatype, Table):
            fastcs_datatypes = [
                numpy_to_fastcs_datatype(datatype)
                for _, datatype in fastcs_datatype.structured_dtype
            ]

            base_get_read_widget = super()._get_read_widget
            widgets = [base_get_read_widget(datatype) for datatype in fastcs_datatypes]

            return TableRead(widgets=widgets)  # type: ignore
        else:
            return super()._get_read_widget(fastcs_datatype)

    def _get_write_widget(self, fastcs_datatype: DataType) -> WriteWidgetUnion | None:
        if isinstance(fastcs_datatype, Table):
            widgets = []
            for _, datatype in fastcs_datatype.structured_dtype:
                fastcs_datatype = numpy_to_fastcs_datatype(datatype)
                if isinstance(fastcs_datatype, Bool):
                    # Replace with compact version for Table row
                    widget = CheckBox()
                else:
                    widget = super()._get_write_widget(fastcs_datatype)
                widgets.append(widget)
            return TableWrite(widgets=widgets)
        else:
            return super()._get_write_widget(fastcs_datatype)



from typing import Any, Generic, cast, get_args

from fastcs.attribute_io_ref import AttributeIORef, AttributeIORefT
from fastcs.attributes import AttrR, AttrW
from fastcs.datatypes import T
from fastcs.tracer import Tracer


class AttributeIO(Generic[T, AttributeIORefT], Tracer):
    """Base class for performing IO for an ``Attribute``

    This class should be inherited to implement reading and writing values from
    ``Attributes`` via some API. For read, ``Attribute``s implement the ``update``
    method and for write, ``Attribute`` implement the ``send`` method.

    Concrete implementations of this class must be parameterised with a specific
    ``AttributeIORef`` that defines exactly what part of the API the ``Attribute``
    corresponds to. See the docstring for ``AttributeIORef`` for more information.
    """

    ref_type = AttributeIORef

    def __init_subclass__(cls) -> None:
        # sets ref_type from subclass generic args
        # from python 3.12 we can use types.get_original_bases
        args = get_args(cast(Any, cls).__orig_bases__[0])
        cls.ref_type = args[1]

    def __init__(self):
        super().__init__()

    async def update(self, attr: AttrR[T, AttributeIORefT]) -> None:
        raise NotImplementedError()

    async def send(self, attr: AttrW[T, AttributeIORefT], value: T) -> None:
        raise NotImplementedError()


AnyAttributeIO = AttributeIO[T, AttributeIORef]



from fastcs.attributes import AttrR
from fastcs.controller import Controller
from fastcs.datatypes import String
from fastcs.launch import FastCS


class TemperatureController(Controller):
    device_id = AttrR(String())


fastcs = FastCS(TemperatureController(), [])

if __name__ == "__main__":
    fastcs.run()



from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.controller import Controller
from fastcs.datatypes import Int
from fastcs.launch import FastCS
from fastcs.transport.epics.ca.transport import EpicsCATransport
from fastcs.transport.epics.options import EpicsIOCOptions
from fastcs.wrappers import command


class ParentController(Controller):
    a: AttrR = AttrR(Int())
    b: AttrRW = AttrRW(Int())


class ChildController(Controller):
    c: AttrW = AttrW(Int())

    @command()
    async def d(self):
        pass


def run(pv_prefix="SOFTIOC_TEST_DEVICE"):
    controller = ParentController()
    controller.child = ChildController()
    fastcs = FastCS(
        controller, [EpicsCATransport(ca_ioc=EpicsIOCOptions(pv_prefix=pv_prefix))]
    )
    fastcs.run(interactive=False)


if __name__ == "__main__":
    run()



import asyncio
from collections.abc import Callable
from typing import Any, Literal

from softioc import builder, softioc
from softioc.asyncio_dispatcher import AsyncioDispatcher
from softioc.pythonSoftIoc import RecordWrapper

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.controller_api import ControllerAPI
from fastcs.cs_methods import Command
from fastcs.datatypes import DataType, T
from fastcs.logging import logger as _logger
from fastcs.tracer import Tracer
from fastcs.transport.epics.ca.util import (
    builder_callable_from_attribute,
    cast_from_epics_type,
    cast_to_epics_type,
    record_metadata_from_attribute,
    record_metadata_from_datatype,
)
from fastcs.transport.epics.options import EpicsIOCOptions
from fastcs.transport.epics.util import controller_pv_prefix
from fastcs.util import snake_to_pascal

EPICS_MAX_NAME_LENGTH = 60


tracer = Tracer(name=__name__)
logger = _logger.bind(logger_name=__name__)


class EpicsCAIOC:
    """A softioc which handles a controller"""

    def __init__(
        self,
        pv_prefix: str,
        controller_api: ControllerAPI,
        options: EpicsIOCOptions | None = None,
    ):
        self._options = options or EpicsIOCOptions()
        self._controller_api = controller_api
        _add_pvi_info(f"{pv_prefix}:PVI")
        _add_sub_controller_pvi_info(pv_prefix, controller_api)

        _create_and_link_attribute_pvs(pv_prefix, controller_api)
        _create_and_link_command_pvs(pv_prefix, controller_api)

    def run(
        self,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        dispatcher = AsyncioDispatcher(loop)  # Needs running loop
        builder.LoadDatabase()
        softioc.iocInit(dispatcher)


def _add_pvi_info(
    pvi: str,
    parent_pvi: str = "",
    name: str = "",
):
    """Add PVI metadata for a controller.

    Args:
        pvi: PVI PV of controller
        parent_pvi: PVI PV of parent controller
        name: Name to register controller with parent as

    """
    # Create a record to attach the info tags to
    record = builder.longStringIn(
        f"{pvi}_PV",
        initial_value=pvi,
        DESC="The records in this controller",
    )

    # Create PVI PV in preparation for adding attribute info tags to it
    q_group = {
        pvi: {
            "+id": "epics:nt/NTPVI:1.0",
            "display.description": {"+type": "plain", "+channel": "DESC"},
            "": {"+type": "meta", "+channel": "VAL"},
        }
    }
    # If this controller has a parent, add a link in the parent to this controller
    if parent_pvi and name:
        q_group.update(
            {
                parent_pvi: {
                    f"value.{name}.d": {
                        "+channel": "VAL",
                        "+type": "plain",
                        "+trigger": f"value.{name}.d",
                    }
                }
            }
        )

    record.add_info("Q:group", q_group)


def _add_sub_controller_pvi_info(pv_prefix: str, parent: ControllerAPI):
    """Add PVI references from controller to its sub controllers, recursively.

    Args:
        pv_prefix: PV Prefix of IOC
        parent: Controller to add PVI refs for

    """
    parent_pvi = f"{controller_pv_prefix(pv_prefix, parent)}:PVI"

    for child in parent.sub_apis.values():
        child_pvi = f"{controller_pv_prefix(pv_prefix, child)}:PVI"
        child_name = child.path[-1].lower()

        _add_pvi_info(child_pvi, parent_pvi, child_name)
        _add_sub_controller_pvi_info(pv_prefix, child)


def _create_and_link_attribute_pvs(
    root_pv_prefix: str, root_controller_api: ControllerAPI
) -> None:
    for controller_api in root_controller_api.walk_api():
        pv_prefix = controller_pv_prefix(root_pv_prefix, controller_api)

        for attr_name, attribute in controller_api.attributes.items():
            pv_name = snake_to_pascal(attr_name)

            full_pv_name_length = len(f"{pv_prefix}:{pv_name}")
            if full_pv_name_length > EPICS_MAX_NAME_LENGTH:
                attribute.enabled = False
                print(
                    f"Not creating PV for {attr_name} for controller"
                    f" {controller_api.path} as full name would exceed"
                    f" {EPICS_MAX_NAME_LENGTH} characters"
                )
                continue

            match attribute:
                case AttrRW():
                    if full_pv_name_length > (EPICS_MAX_NAME_LENGTH - 4):
                        print(
                            f"Not creating PVs for {attr_name} as _RBV PV"
                            f" name would exceed {EPICS_MAX_NAME_LENGTH}"
                            " characters"
                        )
                        attribute.enabled = False
                    else:
                        _create_and_link_read_pv(
                            pv_prefix, f"{pv_name}_RBV", attr_name, attribute
                        )
                        _create_and_link_write_pv(
                            pv_prefix, pv_name, attr_name, attribute
                        )
                case AttrR():
                    _create_and_link_read_pv(pv_prefix, pv_name, attr_name, attribute)
                case AttrW():
                    _create_and_link_write_pv(pv_prefix, pv_name, attr_name, attribute)


def _create_and_link_read_pv(
    pv_prefix: str, pv_name: str, attr_name: str, attribute: AttrR[T]
) -> None:
    pv = f"{pv_prefix}:{pv_name}"

    async def async_record_set(value: T):
        tracer.log_event("PV set from attribute", topic=attribute, pv=pv, value=value)

        record.set(cast_to_epics_type(attribute.datatype, value))

    record = _make_record(pv, attribute)
    _add_attr_pvi_info(record, pv_prefix, attr_name, "r")

    attribute.add_on_update_callback(async_record_set)


def _make_record(
    pv: str,
    attribute: AttrR | AttrW | AttrRW,
    on_update: Callable | None = None,
    out_record: bool = False,
) -> RecordWrapper:
    builder_callable = builder_callable_from_attribute(attribute, on_update is None)
    datatype_record_metadata = record_metadata_from_datatype(
        attribute.datatype, out_record
    )
    attribute_record_metadata = record_metadata_from_attribute(attribute)

    update = {"always_update": True, "on_update": on_update} if on_update else {}

    record = builder_callable(
        pv, **update, **datatype_record_metadata, **attribute_record_metadata
    )

    def datatype_updater(datatype: DataType):
        for name, value in record_metadata_from_datatype(datatype, out_record).items():
            record.set_field(name, value)

    attribute.add_update_datatype_callback(datatype_updater)
    return record


def _create_and_link_write_pv(
    pv_prefix: str, pv_name: str, attr_name: str, attribute: AttrW[T]
) -> None:
    pv = f"{pv_prefix}:{pv_name}"

    async def on_update(value):
        logger.info("PV put: {pv} = {value}", pv=pv, value=value)

        await attribute.put(cast_from_epics_type(attribute.datatype, value))

    async def set_setpoint_without_process(value: T):
        tracer.log_event(
            "PV setpoint set from attribute", topic=attribute, pv=pv, value=value
        )

        record.set(cast_to_epics_type(attribute.datatype, value), process=False)

    record = _make_record(pv, attribute, on_update=on_update, out_record=True)

    _add_attr_pvi_info(record, pv_prefix, attr_name, "w")

    attribute.add_sync_setpoint_callback(set_setpoint_without_process)


def _create_and_link_command_pvs(
    root_pv_prefix: str, root_controller_api: ControllerAPI
) -> None:
    for controller_api in root_controller_api.walk_api():
        pv_prefix = controller_pv_prefix(root_pv_prefix, controller_api)

        for attr_name, method in controller_api.command_methods.items():
            pv_name = snake_to_pascal(attr_name)

            if len(f"{pv_prefix}:{pv_name}") > EPICS_MAX_NAME_LENGTH:
                print(
                    f"Not creating PV for {attr_name} as full name would exceed"
                    f" {EPICS_MAX_NAME_LENGTH} characters"
                )
                method.enabled = False
            else:
                _create_and_link_command_pv(
                    pv_prefix,
                    pv_name,
                    attr_name,
                    method,
                )


def _create_and_link_command_pv(
    pv_prefix: str, pv_name: str, attr_name: str, method: Command
) -> None:
    pv = f"{pv_prefix}:{pv_name}"

    async def wrapped_method(_: Any):
        tracer.log_event("Command PV put", topic=method, pv=pv)

        await method.fn()

    record = builder.Action(
        f"{pv_prefix}:{pv_name}",
        on_update=wrapped_method,
        blocking=True,
    )

    _add_attr_pvi_info(record, pv_prefix, attr_name, "x")


def _add_attr_pvi_info(
    record: RecordWrapper,
    prefix: str,
    name: str,
    access_mode: Literal["r", "w", "rw", "x"],
):
    """Add an info tag to a record to include it in the PVI for the controller.

    Args:
        record: Record to add info tag to
        prefix: PV prefix of controller
        name: Name of parameter to add to PVI
        access_mode: Access mode of parameter

    """
    record.add_info(
        "Q:group",
        {
            f"{prefix}:PVI": {
                f"value.{name}.{access_mode}": {
                    "+channel": "NAME",
                    "+type": "plain",
                    "+trigger": f"value.{name}.{access_mode}",
                }
            }
        },
    )



from __future__ import annotations

import enum
from abc import abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Generic, TypeVar

import numpy as np
from numpy.typing import DTypeLike

T = TypeVar(
    "T",
    int,  # Int
    float,  # Float
    bool,  # Bool
    str,  # String
    enum.Enum,  # Enum
    np.ndarray,  # Waveform
    list[tuple[str, DTypeLike]],  # Table
)

ATTRIBUTE_TYPES: tuple[type] = T.__constraints__  # type: ignore


@dataclass(frozen=True)
class DataType(Generic[T]):
    """Generic datatype mapping to a python type, with additional metadata."""

    @property
    @abstractmethod
    def dtype(self) -> type[T]:  # Using property due to lack of Generic ClassVars
        pass

    def validate(self, value: Any) -> T:
        """Validate a value against the datatype.

        The base implementation is to try the cast and raise a useful error if it fails.

        Child classes can implement logic before calling ``super.validate(value)`` to
        modify the value passed in and help the cast succeed or after to perform further
        validation of the coerced type.

        """
        if isinstance(value, self.dtype):
            return value

        try:
            return self.dtype(value)
        except (ValueError, TypeError) as e:
            raise ValueError(f"Failed to cast {value} to type {self.dtype}") from e

    @property
    @abstractmethod
    def initial_value(self) -> T:
        pass


T_Numerical = TypeVar("T_Numerical", int, float)


@dataclass(frozen=True)
class _Numerical(DataType[T_Numerical]):
    units: str | None = None
    min: T_Numerical | None = None
    max: T_Numerical | None = None
    min_alarm: T_Numerical | None = None
    max_alarm: T_Numerical | None = None

    def validate(self, value: Any) -> T_Numerical:
        _value = super().validate(value)

        if self.min is not None and _value < self.min:
            raise ValueError(f"Value {_value} is less than minimum {self.min}")

        if self.max is not None and _value > self.max:
            raise ValueError(f"Value {_value} is greater than maximum {self.max}")

        return _value

    @property
    def initial_value(self) -> T_Numerical:
        return self.dtype(0)


@dataclass(frozen=True)
class Int(_Numerical[int]):
    """`DataType` mapping to builtin ``int``."""

    @property
    def dtype(self) -> type[int]:
        return int


@dataclass(frozen=True)
class Float(_Numerical[float]):
    """`DataType` mapping to builtin ``float``."""

    prec: int = 2

    @property
    def dtype(self) -> type[float]:
        return float

    def validate(self, value: Any) -> float:
        _value = super().validate(value)

        if self.prec is not None:
            _value = round(_value, self.prec)

        return _value


@dataclass(frozen=True)
class Bool(DataType[bool]):
    """`DataType` mapping to builtin ``bool``."""

    @property
    def dtype(self) -> type[bool]:
        return bool

    @property
    def initial_value(self) -> bool:
        return False


@dataclass(frozen=True)
class String(DataType[str]):
    """`DataType` mapping to builtin ``str``."""

    @property
    def dtype(self) -> type[str]:
        return str

    @property
    def initial_value(self) -> str:
        return ""


T_Enum = TypeVar("T_Enum", bound=enum.Enum)


@dataclass(frozen=True)
class Enum(Generic[T_Enum], DataType[T_Enum]):
    enum_cls: type[T_Enum]

    def __post_init__(self):
        if not issubclass(self.enum_cls, enum.Enum):
            raise ValueError("Enum class has to take an Enum.")

    def index_of(self, value: T_Enum) -> int:
        return self.members.index(value)

    @cached_property
    def members(self) -> list[T_Enum]:
        return list(self.enum_cls)

    @cached_property
    def names(self) -> list[str]:
        return [member.name for member in self.members]

    @property
    def dtype(self) -> type[T_Enum]:
        return self.enum_cls

    @property
    def initial_value(self) -> T_Enum:
        return self.members[0]


@dataclass(frozen=True)
class Waveform(DataType[np.ndarray]):
    array_dtype: DTypeLike
    shape: tuple[int, ...] = (2000,)

    @property
    def dtype(self) -> type[np.ndarray]:
        return np.ndarray

    @property
    def initial_value(self) -> np.ndarray:
        return np.zeros(self.shape, dtype=self.array_dtype)

    def validate(self, value: np.ndarray) -> np.ndarray:
        _value = super().validate(value)

        if self.array_dtype != _value.dtype:
            raise ValueError(
                f"Value dtype {_value.dtype} is not the same as the array dtype "
                f"{self.array_dtype}"
            )

        if len(self.shape) != len(_value.shape) or any(
            shape1 > shape2
            for shape1, shape2 in zip(_value.shape, self.shape, strict=True)
        ):
            raise ValueError(
                f"Value shape {_value.shape} exceeeds the shape maximum shape "
                f"{self.shape}"
            )

        return _value


@dataclass(frozen=True)
class Table(DataType[np.ndarray]):
    # https://numpy.org/devdocs/user/basics.rec.html#structured-datatype-creation
    structured_dtype: list[tuple[str, DTypeLike]]

    @property
    def dtype(self) -> type[np.ndarray]:
        return np.ndarray

    @property
    def initial_value(self) -> np.ndarray:
        return np.array([], dtype=self.structured_dtype)

    def validate(self, value: Any) -> np.ndarray:
        _value = super().validate(value)

        if self.structured_dtype != _value.dtype:
            raise ValueError(
                f"Value dtype {_value.dtype.descr} is not the same as the structured "
                f"dtype {self.structured_dtype}"
            )

        return _value



from pathlib import Path

from fastcs.attributes import AttrR
from fastcs.controller import Controller
from fastcs.datatypes import String
from fastcs.launch import FastCS
from fastcs.transport.epics.ca.options import EpicsGUIOptions
from fastcs.transport.epics.ca.transport import EpicsCATransport
from fastcs.transport.epics.options import EpicsIOCOptions


class TemperatureController(Controller):
    device_id = AttrR(String())


gui_options = EpicsGUIOptions(
    output_path=Path(".") / "demo.bob", title="Demo Temperature Controller"
)
epics_ca = EpicsCATransport(gui=gui_options, ca_ioc=EpicsIOCOptions(pv_prefix="DEMO"))
fastcs = FastCS(TemperatureController(), [epics_ca])

if __name__ == "__main__":
    fastcs.run()



from fastcs.attributes import AttrR
from fastcs.controller import Controller
from fastcs.datatypes import String
from fastcs.launch import FastCS
from fastcs.transport.epics.ca.transport import EpicsCATransport
from fastcs.transport.epics.options import EpicsIOCOptions


class TemperatureController(Controller):
    device_id = AttrR(String())


epics_ca = EpicsCATransport(ca_ioc=EpicsIOCOptions(pv_prefix="DEMO"))
fastcs = FastCS(TemperatureController(), [epics_ca])

if __name__ == "__main__":
    fastcs.run()



from collections.abc import Callable, Coroutine
from typing import Any

import uvicorn
from fastapi import FastAPI
from pydantic import create_model

from fastcs.attributes import AttrR, AttrRW, AttrW, T
from fastcs.controller_api import ControllerAPI
from fastcs.cs_methods import CommandCallback
from fastcs.logging import intercept_std_logger

from .options import RestServerOptions
from .util import (
    cast_from_rest_type,
    cast_to_rest_type,
    convert_datatype,
)


class RestServer:
    """A Rest Server which handles a controller"""

    def __init__(self, controller_api: ControllerAPI):
        self._controller_api = controller_api
        self._app = self._create_app()

    def _create_app(self):
        app = FastAPI()
        _add_attribute_api_routes(app, self._controller_api)
        _add_command_api_routes(app, self._controller_api)

        return app

    async def serve(self, options: RestServerOptions | None):
        options = options or RestServerOptions()
        self._server = uvicorn.Server(
            uvicorn.Config(
                app=self._app,
                host=options.host,
                port=options.port,
                log_level=options.log_level,
            )
        )
        intercept_std_logger("uvicorn.error")
        intercept_std_logger("uvicorn.access")
        intercept_std_logger("uvicorn.asgi")

        await self._server.serve()


def _put_request_body(attribute: AttrW[T]):
    """
    Creates a pydantic model for each datatype which defines the schema
    of the PUT request body
    """
    converted_datatype = convert_datatype(attribute.datatype)
    type_name = str(attribute.datatype.dtype.__name__).title()
    # key=(type, ...) to declare a field without default value
    return create_model(
        f"Put{type_name}Value",
        value=(converted_datatype, ...),
    )


def _wrap_attr_put(
    attribute: AttrW[T],
) -> Callable[[T], Coroutine[Any, Any, None]]:
    async def attr_put(request):
        await attribute.put(cast_from_rest_type(attribute.datatype, request.value))

    # Fast api uses type annotations for validation, schema, conversions
    attr_put.__annotations__["request"] = _put_request_body(attribute)

    return attr_put


def _get_response_body(attribute: AttrR[T]):
    """
    Creates a pydantic model for each datatype which defines the schema
    of the GET request body
    """
    converted_datatype = convert_datatype(attribute.datatype)
    type_name = str(converted_datatype.__name__).title()
    # key=(type, ...) to declare a field without default value
    return create_model(
        f"Get{type_name}Value",
        value=(converted_datatype, ...),
    )


def _wrap_attr_get(
    attribute: AttrR[T],
) -> Callable[[], Coroutine[Any, Any, Any]]:
    async def attr_get() -> Any:  # Must be any as response_model is set
        value = attribute.get()  # type: ignore
        return {"value": cast_to_rest_type(attribute.datatype, value)}

    return attr_get


def _add_attribute_api_routes(app: FastAPI, root_controller_api: ControllerAPI) -> None:
    for controller_api in root_controller_api.walk_api():
        path = controller_api.path

        for attr_name, attribute in controller_api.attributes.items():
            attr_name = attr_name.replace("_", "-")
            route = f"{'/'.join(path)}/{attr_name}" if path else attr_name

            match attribute:
                # https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods
                case AttrRW():
                    app.add_api_route(
                        f"/{route}",
                        _wrap_attr_get(attribute),
                        methods=["GET"],  # Idempotent and safe data retrieval,
                        status_code=200,  # https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/GET
                        response_model=_get_response_body(attribute),
                    )
                    app.add_api_route(
                        f"/{route}",
                        _wrap_attr_put(attribute),
                        methods=["PUT"],  # Idempotent state change
                        status_code=204,  # https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/PUT
                    )
                case AttrR():
                    app.add_api_route(
                        f"/{route}",
                        _wrap_attr_get(attribute),
                        methods=["GET"],
                        status_code=200,
                        response_model=_get_response_body(attribute),
                    )
                case AttrW():
                    app.add_api_route(
                        f"/{route}",
                        _wrap_attr_put(attribute),
                        methods=["PUT"],
                        status_code=204,
                    )


def _wrap_command(
    method: CommandCallback,
) -> Callable[..., Coroutine[None, None, None]]:
    async def command() -> None:
        await method()

    return command


def _add_command_api_routes(app: FastAPI, root_controller_api: ControllerAPI) -> None:
    for controller_api in root_controller_api.walk_api():
        path = controller_api.path

        for name, method in root_controller_api.command_methods.items():
            cmd_name = name.replace("_", "-")
            route = f"/{'/'.join(path)}/{cmd_name}" if path else cmd_name
            app.add_api_route(
                f"/{route}",
                _wrap_command(method.fn),
                methods=["PUT"],
                status_code=204,
            )



from .transport import Transport as Transport

try:
    from .epics.ca.transport import EpicsCATransport as EpicsCATransport
    from .epics.options import EpicsDocsOptions as EpicsDocsOptions
    from .epics.options import EpicsGUIOptions as EpicsGUIOptions
    from .epics.options import EpicsIOCOptions as EpicsIOCOptions
except ImportError:
    pass

try:
    from .epics.pva.transport import EpicsPVATransport as EpicsPVATransport
except ImportError:
    pass

try:
    from .graphql.transport import GraphQLTransport as GraphQLTransport
except ImportError:
    pass

try:
    from .rest.transport import RestTransport as RestTransport
except ImportError:
    pass

try:
    from .tango.options import TangoDSROptions as TangoDSROptions
    from .tango.transport import TangoTransport as TangoTransport
except ImportError:
    pass
